package PolicyEngine;

use XML::LibXML;
use Time::HiRes qw(gettimeofday);
use SecureComponent;

#TODO - Could check to see if these files exist before trying to read them
my $POLICY_FILE_PATH = 'config/policies.xml';
my $POLICY_PREFERENCES_FILE_PATH = 'config/policyPreferences.xml';



my $PRINT_INFORMATION = 0;
my $PRINT_ERROR = 1;
my $PRINT_SUMMARY = 0;

my $MODULE_NAME = 'PolicyEngine';
my $PRINT_INFORMATION_PREFIX = 'INFORMATION - ' . $MODULE_NAME . ' - ';
my $PRINT_WARNING_PREFIX = 'WARNING - ' . $MODULE_NAME . ' - ';
my $PRINT_ERROR_PREFIX = 'FATAL ERROR - ' . $MODULE_NAME . ' - ';
my $PRINT_LOCAL_PREFIX = '';

my $MEASURING_COMPONENTS = 1;



my $ANP_NAMESPACE = 'urn:ucb:eecs:sahara:sso:ANP:1.0';

my $AUTH_METHOD_ELEMENT_NAME = 'AuthnMethod';
my $AUTH_METHOD_ID_ELEMENT_NAME = 'AuthnMethodID';

my $CHARGING_OPTION_ELEMENT_NAME = 'ChargingOption';
my $CHARGING_OPTION_ID_ELEMENT_NAME = 'ChargingOptionID';

my $TIME_UNIT_ELEMENT_NAME = 'TimeUnit';
my $UNIT_PRICE_ELEMENT_NAME = 'UnitPrice';

my $PREFERENCE_GROUP_ELEMENT_NAME = 'PreferenceGroup';
my $WEIGHTING_EQUATION_ELEMENT_NAME = 'WeightingEquation';

my $IDP_GROUP_ELEMENT_NAME = 'IDPGroup';
my $IDP_LIST_ELEMENT_NAME = 'IDPList';
my $IDP_NAME_ELEMENT_NAME = 'IDPName';
my $IDP_DOMAIN_ELEMENT_NAME = 'IDPDomain';

my $AUTH_METHOD_ID_REFERENCE_ELEMENT_NAME = 'AuthnMethodIDReference';
my $CHARGING_OPTION_ID_REFERENCE_ELEMENT_NAME = 'ChargingOptionIDReference';

my $USER_INFO_DESIGNATOR_ELEMENT_NAME = 'UserInfoDesignator';

my $RULE_ELEMENT_NAME = 'rule';
my $SUBJECT_ELEMENT_NAME = 'subject';
my $ATTRIBUTE_ELEMENT_NAME = 'attribute';
my $AUTH_INFO_ELEMENT_NAME = 'authn_info';
my $PROVISIONAL_ACTION_ELEMENT_NAME = 'provisional_action';

my $PREFERENCE_GROUP_MEMBER_ELEMENT_NAME = 'GroupMember';

my $NONCE_ELEMENT_NAME = 'Nonce';



my $TYPE_ATTRIBUTE_NAME = 'type';
my $EQUATION_ATTRIBUTE_NAME = 'equation';

my $HREF_ATTRIBUTE_NAME = 'href';
my $NAME_ATTRIBUTE_NAME = 'name';

my $ATTRIBUTE_NAME_ATTRIBUTE_NAME = 'AttributeName';

my $GROUP_MEMBER_NAME_ATTRIBUTE_NAME = 'name';
my $GROUP_MEMBER_WEIGHT_ATTRIBUTE_NAME = 'weight';



my $INFO_DESIGNATOR_ELEMENTS_KEY = 'infoDesignatorElements';
my $WEIGHT_KEY = 'weight';
my $TIME_UNIT_KEY = 'timeUnit';
my $UNIT_PRICE_KEY = 'unitPrice';
my $ID_PROVIDER_NAME_KEY = 'IDProvider';
my $AUTH_METHOD_ID_KEY = 'authMethod';
my $CHARGING_OPTION_ID_KEY = 'chargingOption';

my $SECURE_INFO_ACCESSOR_KEY = 'secureInfoAccessor';

my $MEASUREMENT_DESCRIPTIONS_KEY = 'measurementDescriptions';
my $MEASUREMENT_TIMES_KEY = 'measurementTimes';



my $AUTH_METHOD_PREFERENCE_TYPE = 'AuthMethod';
my $CHARGING_OPTION_PREFERENCE_TYPE = 'ChargingOption';
my $ID_PROVIDER_PREFERENCE_TYPE = 'IDProvider';


my $LINEAR_EQUATION_NAME = 'linear';
my $NULL_EQUATION_NAME = 'none';



my $TRUSTED_TRUST_LEVEL = 'trusted';
my $UNTRUSTED_TRUST_LEVEL = 'untrusted';
my $DEFAULT_TRUST_LEVEL = 'default';



my $USER_ACKNOWLEDGEMENT_ACTION_NAME = 'user_ack';
my $USER_NOTIFICATION_ACTION_NAME = 'user_notification';
my $USER_INPUT_ACTION_NAME = 'user_input';



my $UNUSABLE_TRIPLET_WEIGHT = 0;

my $ZERO_WEIGHT = 0;



my $TRUE = 1;
my $FALSE = 0;

my $NORMAL_RETURN_STATUS = $TRUE;
my $ERROR_RETURN_STATUS = $FALSE;



sub new
{
    my $PRINT_LOCAL_PREFIX = 'new() - ';

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Creating new policy engine.\n");
    }

    my $class = shift;

    my $secureInfoAccessor = SecureComponent->new();
    my $this = { $SECURE_INFO_ACCESSOR_KEY => $secureInfoAccessor };

    bless($this, $class);

    return $this;
}

sub choose
{
    my $PRINT_LOCAL_PREFIX = 'choose() - ';

    my $this = shift;
    my ($capabilitiesStatement, $desiredCapabilitiesStatementNonce) = @_;

    my $measurementDescriptions = [ ];
    my $measurementTimes = [ ];
    if ($MEASURING_COMPONENTS)
    {
	$this->{$MEASUREMENT_DESCRIPTIONS_KEY} = $measurementDescriptions;
	$this->{$MEASUREMENT_TIMES_KEY} = $measurementTimes;

	push(@$measurementDescriptions, "10. Starting choose");
	push(@$measurementTimes, gettimeofday());
    }    

    my $secureInfoAccessor = $this->{$SECURE_INFO_ACCESSOR_KEY};

    my $XMLParser = XML::LibXML->new();

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Parsing xml files.\n");
    }

    #TODO - Drop in eval blocks and check for parsing errors
    my $capabilitiesStatementParseTree = $XMLParser->parse_string($capabilitiesStatement);
    my $policyParseTree = $XMLParser->parse_file($POLICY_FILE_PATH);
    my $policyPreferencesParseTree = $XMLParser->parse_file($POLICY_PREFERENCES_FILE_PATH);

    if ($MEASURING_COMPONENTS)
    {
	push(@$measurementDescriptions, "20. Done parsing capabilities statement, policies, and policy preferences");
	push(@$measurementTimes, gettimeofday());
    }

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Gathering info.\n");
    }

    #TODO - Comment that these are hash refs
    my $authMethodInfo = $this->gatherAuthMethodInfo($capabilitiesStatementParseTree, $policyPreferencesParseTree);
    if ($MEASURING_COMPONENTS)
    {
	push(@$measurementDescriptions, "30. Done gathering auth method info");
	push(@$measurementTimes, gettimeofday());
    }

    my $chargingOptionInfo = $this->gatherChargingOptionInfo($capabilitiesStatementParseTree,
							     $policyPreferencesParseTree);
    if ($MEASURING_COMPONENTS)
    {
	push(@$measurementDescriptions, "40. Done gathering charging info");
	push(@$measurementTimes, gettimeofday());
    }
    my $IDProviderInfo = $this->gatherIDProviderInfo($capabilitiesStatementParseTree, $policyPreferencesParseTree);
    if ($MEASURING_COMPONENTS)
    {
	push(@$measurementDescriptions, "50. Done gathering ID provider info");
	push(@$measurementTimes, gettimeofday());
    }

    #TODO - This will be an array reference. Put a comment noting this
    my $tripletCombinationInfo = $this
	->computeTripletCombinationInfo($capabilitiesStatementParseTree,
					$authMethodInfo, $chargingOptionInfo, $IDProviderInfo);
    if ($MEASURING_COMPONENTS)
    {
	push(@$measurementDescriptions, "60. Done computing triplet info");
	push(@$measurementTimes, gettimeofday());
    }

    my $trustLevel = $this->determineTrustLevel($capabilitiesStatementParseTree, $desiredCapabilitiesStatementNonce);
    if ($MEASURING_COMPONENTS)
    {
	push(@$measurementDescriptions, "70. Done determining trust level");
	push(@$measurementTimes, gettimeofday());
    }

    my $chosenTripletInfo = $this->chooseBestAllowedTriplet($policyParseTree, $tripletCombinationInfo, $trustLevel);
    if ($MEASURING_COMPONENTS)
    {
	push(@$measurementDescriptions, "80. Done choosing best triplet at trusted/untrusted trust level");
	push(@$measurementTimes, gettimeofday());
    }

    if (!($chosenTripletInfo))
    {
	$trustLevel = $DEFAULT_TRUST_LEVEL;

	$chosenTripletInfo = $this->chooseBestAllowedTriplet($policyParseTree, $tripletCombinationInfo, $trustLevel);
	if ($MEASURING_COMPONENTS)
	{
	    push(@$measurementDescriptions, "90. Done choosing best triplet at default trust level");
	    push(@$measurementTimes, gettimeofday());
	}

	if (!($chosenTripletInfo))
	{
	    if ($PRINT_ERROR)
	    {
		print(STDERR $PRINT_ERROR_PREFIX . $PRINT_LOCAL_PREFIX .
		      "No acceptable triplet was found!.\n");

		if ($MEASURING_COMPONENTS)
		{
		    push(@$measurementDescriptions, "100. Exiting");
		    push(@$measurementTimes, gettimeofday());
		}

		return $ERROR_RETURN_STATUS;
	    }
	}
    }

    my $provisionalActionResult = $this->executeProvisionalAction($policyParseTree, $trustLevel);
    if ($MEASURING_COMPONENTS)
    {
	push(@$measurementDescriptions, "110. Done executing provisional action");
	push(@$measurementTimes, gettimeofday());
    }

    if (!($provisionalActionResult))
    {
	#TODO - Still need to print an error
	if ($MEASURING_COMPONENTS)
	{
	    push(@$measurementDescriptions, "120. Exiting");
	    push(@$measurementTimes, gettimeofday());
	}

	return $ERROR_RETURN_STATUS;
    }
    
    ($chosenTripletInfo->{$ID_PROVIDER_NAME_KEY}) = split(/_/, $chosenTripletInfo->{$ID_PROVIDER_NAME_KEY});

    my ($requiredInfoDesignatorNames, $requiredInfoDesignatorValues) =
	$this->gatherSecureInfo($secureInfoAccessor,
				$chosenTripletInfo->{$INFO_DESIGNATOR_ELEMENTS_KEY},
				$chosenTripletInfo->{$ID_PROVIDER_NAME_KEY});

    if ($MEASURING_COMPONENTS)
    {
	push(@$measurementDescriptions, "130. Done gathering secure info");
	push(@$measurementTimes, gettimeofday());
    }


    if ($PRINT_SUMMARY)
    {
	$this->printAuthMethodInfo($authMethodInfo);
	$this->printChargingOptionInfo($chargingOptionInfo);
	$this->printIDProviderInfo($IDProviderInfo);
	$this->printTripletCombinationInfo($tripletCombinationInfo);
	
	print("===== DECISION INFO =====\n");
	print("  Trust level = \'" . $trustLevel . "\'.\n");
	print("  Chosen triplet = (" . $chosenTripletInfo->{$ID_PROVIDER_NAME_KEY} .
	      ", " . $chosenTripletInfo->{$AUTH_METHOD_ID_KEY} .
	      ", " . $chosenTripletInfo->{$CHARGING_OPTION_ID_KEY} . ").\n");
	print("  Result of provisional action = \'" . $provisionalActionResult . "\'.\n");
    }

    if ($MEASURING_COMPONENTS)
    {
	push(@$measurementDescriptions, "140. Exiting (May include PRINT_SUMMARY time)");
	push(@$measurementTimes, gettimeofday());
    }

    return ($NORMAL_RETURN_STATUS, $chosenTripletInfo->{$ID_PROVIDER_NAME_KEY},
	    $chosenTripletInfo->{$AUTH_METHOD_ID_KEY}, $chosenTripletInfo->{$CHARGING_OPTION_ID_KEY},
	    $requiredInfoDesignatorNames, $requiredInfoDesignatorValues);
}

sub lookupSecureInfo
{
    my $PRINT_LOCAL_PREFIX = 'lookupSecureInfo() - ';

    my $this = shift;
    my ($desiredInfoDesignatorName, $selectedIDProviderDomain) = @_;

    my $secureInfoAccessor = $this->{$SECURE_INFO_ACCESSOR_KEY};

    return ($secureInfoAccessor->lookupSecureInfo($desiredInfoDesignatorName, $selectedIDProviderDomain));
}

sub gatherSecureInfo
{
    my $PRINT_LOCAL_PREFIX = 'gatherSecureConfInfo() - ';

    my $this = shift;
    my ($secureInfoAccessor, $requiredInfoDesignatorElements, $selectedIDProviderName) = @_;

    my $requiredInfoDesignatorNames = $this
	->getInfoDesignatorNamesFromElements($requiredInfoDesignatorElements,
					     $ATTRIBUTE_NAME_ATTRIBUTE_NAME);

    my $requiredInfoDesignatorValues = [ ];

    foreach my $currentInfoDesignatorName (@$requiredInfoDesignatorNames)
    {
	push(@$requiredInfoDesignatorValues, $secureInfoAccessor
	     ->lookupSecureInfo($currentInfoDesignatorName, $selectedIDProviderName));
    }
	
    return ($requiredInfoDesignatorNames, $requiredInfoDesignatorValues);

}

sub executeProvisionalAction
{
    my $PRINT_LOCAL_PREFIX = 'executeProvisionalAction() - ';

    my $this = shift;
    my ($policyParseTree, $trustLevel) = @_;

    #TODO - Need to put in a default message (Probably on a per-action basis).
    my $message = "Provisional action message. Enter y.* if prompted for positive acknowledgement: ";

    my $policyRuleElement = $this->getPolicyRuleElement($policyParseTree, $trustLevel);

    # We will just assume a well formed policy file
    my ($provisionalActionElement) = $policyRuleElement
	->getChildrenByTagName($PROVISIONAL_ACTION_ELEMENT_NAME);

  if ($provisionalActionElement) {
    my $provisionalActionName = $provisionalActionElement
	->getAttribute($NAME_ATTRIBUTE_NAME);

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Executing provisional action with name = \'" . $provisionalActionName . "\'.\n");
    }

    if ($provisionalActionName eq $USER_ACKNOWLEDGEMENT_ACTION_NAME)
    {
	return $this->executeUserAcknowledgementAction($message);
    }
    elsif ($provisionalActionName eq $USER_NOTIFICATION_ACTION_NAME)
    {
	return $this->executeUserNotificationAction($message);
    }
    elsif ($provisionalActionName eq $USER_INPUT_ACTION)
    {
	return $this->executeUserInputAction($message);
    }
    else
    {
	#TODO - Unknown action. Need to print an error

	return $ERROR_RETURN_STATUS;
    }
  } else {
    return $NORMAL_RETURN_STATUS;
  }
}

sub executeUserAcknowledgementAction
{
    my $this = shift;
    my ($message) = @_;
    
    print($message);
    my $userResponse = readline(STDIN);

    chomp($userResponse);

    if ($userResponse =~ m/y.*/i)
    {
	return $NORMAL_RETURN_STATUS;
    }
    else
    {
	return $ERROR_RETURN_STATUS;
    }
}

sub executeUserNotificationAction
{
    my $this = shift;
    my ($message) = @_;

    chomp($message);
    print($message . "\n");

    return $NORMAL_RETURN_STATUS;
}

sub executeUserInputAction
{
    my $this = shift;
    my ($message) = @_;

    print("TODO\n");

    return $NORMAL_RETURN_STATUS;
}

sub determineTrustLevel
{
    my $this = shift;
    my ($capabilitiesStatementParseTree, $desiredCapabilitiesStatementNonce) = @_;

    if ($desiredCapabilitiesStatementNonce)
    {
	my ($nonceElement) = $capabilitiesStatementParseTree
	    ->getElementsByTagNameNS($ANP_NAMESPACE, $NONCE_ELEMENT_NAME);

	if ($nonceElement)
	{
	    my $capabilitiesStatementNonce = $nonceElement->textContent;

	    if ($capabilitiesStatementNonce ne $desiredCapabilitiesStatementNonce)
	    {
		return $UNTRUSTED_TRUST_LEVEL;
	    }
	}
	else
	{
	    return $UNTRUSTED_TRUST_LEVEL;
	}
    }

    #TODO - Still need to check the signature

    #open(CAPABILITIES_STATEMENT_FILE, '+>', 'capStatement.txt');
    #print(CAPABILITIES_STATEMENT_FILE $capabilitiesStatement);
    #close(CAPABILITIES_STATEMENT_FILE);

    #my $output = `xmlsec1 --verify --pubkey-cert-der liberty.der capStatement.txt`;
    #print("VERIFY OUTPUT = \n" . $output . "\n");

    return $TRUSTED_TRUST_LEVEL;
}

sub chooseBestAllowedTriplet
{
    my $PRINT_LOCAL_PREFIX = 'chooseBestAllowedTriplet() - ';

    my $this = shift;
    my ($policyParseTree, $tripletCombinationInfo, $trustLevel) = @_;

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Looking for triplet at trust level = \'" . $trustLevel . "\'.\n");
    }
    
    my $policyRuleElement = $this->getPolicyRuleElement($policyParseTree, $trustLevel);
    #TODO - Im not going to check the above; If its null, then we are screwed anyways

    my @allowedInfoDesignatorElements = $policyRuleElement->getChildrenByTagName($AUTH_INFO_ELEMENT_NAME);
    
    my $allowedInfoDesignatorNames = $this
	->getInfoDesignatorNamesFromElements(\@allowedInfoDesignatorElements, $HREF_ATTRIBUTE_NAME);

    foreach my $currentTripletInfo (@$tripletCombinationInfo)
    {
	#TODO - Comment (List is ordered, so we will assume we have reached the 0's; More efficient than next;
	if ($currentTripletInfo->{$WEIGHT_KEY} == $UNUSABLE_TRIPLET_WEIGHT)
	{
	    if ($PRINT_INFORMATION)
	    {
		print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		      "Found triplet with 0 weight; Skipping remainder of triplets.\n");
	    }

	    last;
	}

	my $requiredInfoDesignatorNames = $this
	    ->getInfoDesignatorNamesFromElements($currentTripletInfo->{$INFO_DESIGNATOR_ELEMENTS_KEY},
						 $ATTRIBUTE_NAME_ATTRIBUTE_NAME);

	if ($PRINT_INFORMATION)
	{
	    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		  "Considering triplet (" . $currentTripletInfo->{$ID_PROVIDER_NAME_KEY} .
		  ", " . $currentTripletInfo->{$AUTH_METHOD_ID_KEY} .
		  ", " . $currentTripletInfo->{$CHARGING_OPTION_ID_KEY} . ").\n");
	    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		  "Number of designators required = \'" . scalar(@$requiredInfoDesignatorNames) . "\'.\n");
	}

	my $requiresOnlyAllowedInfo = $TRUE;
	#TODO - My own search (Since I cant find a standard function to do it)
	foreach my $currentRequiredDesignatorName (@$requiredInfoDesignatorNames)
	{
	    my $currentNameIsAllowed = $FALSE;

	    #TODO - This is a hack
	    $currentRequiredDesignatorName =~ s/login_info/loginInfo/;
	    $currentRequiredDesignatorName =~ s/credit_card/creditCard/;

	    $currentRequiredDesignatorName =~ s/user_id/userName/;
	    $currentRequiredDesignatorName =~ s/card_number/number/;
	    $currentRequiredDesignatorName =~ s/cardNumber/number/;

	    foreach my $currentAllowedDesignatorName (@$allowedInfoDesignatorNames)
	    {

		if ($PRINT_INFORMATION)
		{
		    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
			  "Required name = \'" . $currentRequiredDesignatorName .
			  "\', current allowed name = \'" . $currentAllowedDesignatorName . "\'.\n");
		}

		if ($currentRequiredDesignatorName eq $currentAllowedDesignatorName)
		{
		    
		    if ($PRINT_INFORMATION)
		    {
			print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
			      "Names match; Current required name is allowed.\n");
		    }

		    $currentNameIsAllowed = $TRUE;
		    last;
		}
	    }

	    if (!($currentNameIsAllowed))
	    {
		if ($PRINT_INFORMATION)
		{
		    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
			  "No match was found; Current required name is not allowed.\n");
		}

		$requiresOnlyAllowedInfo = $FALSE;
		last;
	    }
	}

	if ($requiresOnlyAllowedInfo)
	{
	    if ($PRINT_INFORMATION)
	    {
		print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		      "All required names were allowed; This triplet is a match.\n");
	    }

	    return $currentTripletInfo;
	}
    }

    return;
}

sub getPolicyRuleElement
{
    my $PRINT_LOCAL_PREFIX = 'getPolicyRuleElement() - ';

    my $this = shift;
    my ($policyParseTree, $desiredTrustLevel) = @_;

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Looking for policy rule at trust level = \'" . $desiredTrustLevel . "\'.\n");
    }

    my $desiredPolicyRuleElement;

    my @policyRuleElements = $policyParseTree->getElementsByTagName($RULE_ELEMENT_NAME);

    foreach my $currentPolicyRuleElement (@policyRuleElements)
    {
	#TODO - It might be interesting to rename these in a meaningful way (Describe what they do)
	my ($subjectElement) = $currentPolicyRuleElement->getChildrenByTagName($SUBJECT_ELEMENT_NAME);

	if ($subjectElement)
	{
	    my ($attributeElement) = $subjectElement->getChildrenByTagName($ATTRIBUTE_ELEMENT_NAME);

	    if ($attributeElement)
	    {
		if ($attributeElement->textContent eq $desiredTrustLevel)
		{
		    $desiredPolicyRuleElement = $currentPolicyRuleElement;

		    last;
		}
	    }
	}
    }
    
    #TODO - Put out a warning here if we have a null return value, although this usually isnt something to
    #   worry about; There should be some set of guranteed rules

    return $desiredPolicyRuleElement;
}

sub gatherAuthMethodInfo
{
    my $PRINT_LOCAL_PREFIX = 'gatherAuthMethodInfo() - ';

    my $this = shift;
    my ($capabilitiesStatementParseTree, $policyPreferencesParseTree) = @_;

    my %authMethodInfo;

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Gathering auth method info.\n");
    }

    my $authMethodPreferenceGroupElement =
	$this->getPolicyPreferenceGroup($policyPreferencesParseTree, $AUTH_METHOD_PREFERENCE_TYPE);

    my @authMethodElements = $capabilitiesStatementParseTree
	->getElementsByTagNameNS($ANP_NAMESPACE, $AUTH_METHOD_ELEMENT_NAME);
    
    #TODO - Explain all of the data structures
    foreach my $currentAuthMethodElement (@authMethodElements)
    {
	my %currentAuthMethodInfo;

	#TODO - Going to assume that this is present (Otherwise, the xml isnt well formed (?)).
	my ($authMethodIDElement) = $currentAuthMethodElement
	    ->getChildrenByTagNameNS($ANP_NAMESPACE, $AUTH_METHOD_ID_ELEMENT_NAME);
	my $authMethodID = $authMethodIDElement->textContent;
	
	if ($PRINT_INFORMATION)
	{
	    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		  "Looking at auth method with ID = \'" . $authMethodID . "\'.\n");
	}

	#TODO - Info Designators
	my @infoDesignatorElements = $currentAuthMethodElement
	    ->getChildrenByTagNameNS($ANP_NAMESPACE, $USER_INFO_DESIGNATOR_ELEMENT_NAME);
	$currentAuthMethodInfo{$INFO_DESIGNATOR_ELEMENTS_KEY} = \@infoDesignatorElements;

	my $authMethodWeight =
	    $this->getPreferenceGroupMemberWeight($authMethodPreferenceGroupElement, $authMethodID);
	$currentAuthMethodInfo{$WEIGHT_KEY} = $authMethodWeight;

	$authMethodInfo{$authMethodID} = \%currentAuthMethodInfo;
    }

    return \%authMethodInfo;
}

sub gatherChargingOptionInfo
{
    my $PRINT_LOCAL_PREFIX = 'gatherChargingOptionInfo() - ';

    my $this = shift;
    my ($capabilitiesStatementParseTree, $policyPreferencesParseTree) = @_;

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Gathering charging option info.\n");
    }

    my %chargingOptionInfo;

    #TODO - What this variable actually holds should not be a concern - Only some evaluateEquation func should deal with it
    my $chargingOptionWeightingEquation = 
	$this->getPolicyPreferenceEquation($policyPreferencesParseTree, $CHARGING_OPTION_PREFERENCE_TYPE);

    my @chargingOptionElements = $capabilitiesStatementParseTree
	->getElementsByTagNameNS($ANP_NAMESPACE, $CHARGING_OPTION_ELEMENT_NAME);

    foreach my $currentChargingOptionElement (@chargingOptionElements)
    {
	my %currentChargingOptionInfo;

	#TODO - We will assume that there is an ID element with each ChargingOption
	my ($chargingOptionIDElement) = $currentChargingOptionElement
	    ->getChildrenByTagNameNS($ANP_NAMESPACE, $CHARGING_OPTION_ID_ELEMENT_NAME);
	my $chargingOptionID = $chargingOptionIDElement->textContent;

	if ($PRINT_INFORMATION)
	{
	    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		  "Looking at charging option with ID = \'" . $chargingOptionID . "\'.\n");
	}

	my ($timeUnitElement) = $currentChargingOptionElement
	    ->getChildrenByTagNameNS($ANP_NAMESPACE, $TIME_UNIT_ELEMENT_NAME);
	if ($timeUnitElement)
	{
	    $currentChargingOptionInfo{$TIME_UNIT_KEY} = $timeUnitElement->textContent;
	}

	my ($unitPriceElement) = $currentChargingOptionElement
	    ->getChildrenByTagNameNS($ANP_NAMESPACE, $UNIT_PRICE_ELEMENT_NAME);
	if ($unitPriceElement)
	{
	    $currentChargingOptionInfo{$UNIT_PRICE_KEY} = $unitPriceElement->textContent;
	}

	#TODO - Info designators
	my @infoDesignatorElements = $currentChargingOptionElement
	    ->getChildrenByTagNameNS($ANP_NAMESPACE, $USER_INFO_DESIGNATOR_ELEMENT_NAME);
	$currentChargingOptionInfo{$INFO_DESIGNATOR_ELEMENTS_KEY} = \@infoDesignatorElements;

	$currentChargingOptionInfo{$WEIGHT_KEY} = $this
	    ->evaluateEquation($chargingOptionWeightingEquation, \%currentChargingOptionInfo);

	#TODO - Going to ignore service id reference for now
	#TODO - Everything from now on is optional; It may be null. Keep this in mind later, when retrievin this data.
	#TODO - For now im just gonna get TimeUnit, UnitPrice, and the info designators

	$chargingOptionInfo{$chargingOptionID} = \%currentChargingOptionInfo;
    }
    
    return \%chargingOptionInfo;
}

sub gatherIDProviderInfo
{
    my $PRINT_LOCAL_PREFIX = 'gatherIDProviderInfo() - ';

    my $this = shift;
    my ($capabilitiesStatementParseTree, $policyPreferencesParseTree) = @_;

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Gathering ID Provider info.\n");
    }

    my %IDProviderInfo;

    my @IDPGroupElements = $capabilitiesStatementParseTree
	->getElementsByTagNameNS($ANP_NAMESPACE, $IDP_GROUP_ELEMENT_NAME);

    my $IDProviderPreferenceGroupElement = $this
	->getPolicyPreferenceGroup($policyPreferencesParseTree, $ID_PROVIDER_PREFERENCE_TYPE);

    foreach my $currentIDPGroupElement (@IDPGroupElements)
    {
	my ($IDPListElement) = $currentIDPGroupElement
	    ->getChildrenByTagNameNS($ANP_NAMESPACE, $IDP_LIST_ELEMENT_NAME);

	if ($IDPListElement)
	{
	    my @IDPNameElements = $IDPListElement
		->getChildrenByTagNameNS($ANP_NAMESPACE, $IDP_NAME_ELEMENT_NAME);
	    my @IDPDomainElements = $IDPListElement
		->getChildrenByTagNameNS($ANP_NAMESPACE, $IDP_DOMAIN_ELEMENT_NAME);

	    foreach my $currentIDPNameOrDomainElement (@IDPNameElements, @IDPDomainElements)
	    {
		#TODO - We use this hash, even though its not strictly necessary (We could just have the
		#  name/domain hash to the weight) for future exptensibility.
		my %currentIDProviderInfo;

		my $currentIDPNameOrDomain = $currentIDPNameOrDomainElement->textContent;

		if ($PRINT_INFORMATION)
		{
		    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
			  "Looking at IDP with name or domain = \'" . $currentIDPNameOrDomain . "\'.\n");
		}

		$currentIDProviderInfo{$WEIGHT_KEY} = $this
		    ->getPreferenceGroupMemberWeight($IDProviderPreferenceGroupElement,
						     $currentIDPNameOrDomain);

		$IDProviderInfo{$currentIDPNameOrDomain} = \%currentIDProviderInfo;
	    }
	}
	else
	{
	    if ($PRINT_INFORMATION)
	    {
		print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		      "IDPGroup has no associated IDPList.\n");
	    }
	}

    }

    return \%IDProviderInfo;
}

sub computeTripletCombinationInfo
{
    my $PRINT_LOCAL_PREFIX = 'computeTripletCombinationInfo() - ';

    my $this = shift;
    my ($capabilitiesStatementParseTree, $authMethodInfo, $chargingOptionInfo, $IDProviderInfo) = @_;

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Computing triplet combinations and gathering associated info.\n");
    }

    my @tripletCombinationInfo;

    my @IDPGroupElements = $capabilitiesStatementParseTree
	->getElementsByTagNameNS($ANP_NAMESPACE, $IDP_GROUP_ELEMENT_NAME);

    foreach my $currentIDPGroupElement (@IDPGroupElements)
    {
	my ($IDPListElement) = $currentIDPGroupElement
	    ->getChildrenByTagNameNS($ANP_NAMESPACE, $IDP_LIST_ELEMENT_NAME);

	if ($IDPListElement)
	{
	    my @IDPNameElements = $IDPListElement
		->getChildrenByTagNameNS($ANP_NAMESPACE, $IDP_NAME_ELEMENT_NAME);
	    my @IDPDomainElements = $IDPListElement
		->getChildrenByTagNameNS($ANP_NAMESPACE, $IDP_DOMAIN_ELEMENT_NAME);

	    my @authMethodIDElements = $currentIDPGroupElement
		->getChildrenByTagNameNS($ANP_NAMESPACE, $AUTH_METHOD_ID_REFERENCE_ELEMENT_NAME);
					 
	    my @chargingOptionIDElements = $currentIDPGroupElement
		->getChildrenByTagNameNS($ANP_NAMESPACE, $CHARGING_OPTION_ID_REFERENCE_ELEMENT_NAME);

	    #TODO - Possibly rearange these so that they are more efficient

	    #For now, im just gonna keep track of the required info designator elements

	    foreach my $currentIDPNameOrDomainElement (@IDPNameElements, @IDPDomainElements)
	    {
		my $currentIDPNameOrDomain = $currentIDPNameOrDomainElement->textContent;
		my $currentIDProviderInfo = $IDProviderInfo->{$currentIDPNameOrDomain};
		my $currentIDProviderWeight = $currentIDProviderInfo->{$WEIGHT_KEY};

		foreach my $currentAuthMethodIDElement (@authMethodIDElements)
		{
		    my $currentAuthMethodID = $currentAuthMethodIDElement->textContent;
		    my $currentAuthMethodInfo = $authMethodInfo->{$currentAuthMethodID};
		    my $currentAuthMethodWeight = $currentAuthMethodInfo->{$WEIGHT_KEY};
		    my $authMethodInfoDesignatorElements =
			$currentAuthMethodInfo->{$INFO_DESIGNATOR_ELEMENTS_KEY};

		    foreach my $currentChargingOptionIDElement (@chargingOptionIDElements)
		    {
			my $currentChargingOptionID = $currentChargingOptionIDElement->textContent;
			my $currentChargingOptionInfo = $chargingOptionInfo->{$currentChargingOptionID};
			my $currentChargingOptionWeight = $currentChargingOptionInfo->{$WEIGHT_KEY};
			my $chargingOptionInfoDesignatorElements =
			    $currentChargingOptionInfo->{$INFO_DESIGNATOR_ELEMENTS_KEY};

			#TODO - Ive got major naming issues here (Current or not?)

			my %currentTripletCombinationInfo;

			if ($PRINT_INFORMATION)
			{
			    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
				  "Looking at combo (" . $currentIDPNameOrDomain .
				  ", " . $currentAuthMethodID . ", " . $currentChargingOptionID . ").\n");
			}

			$currentTripletCombinationInfo{$ID_PROVIDER_NAME_KEY} = $currentIDPNameOrDomain;
			$currentTripletCombinationInfo{$AUTH_METHOD_ID_KEY} = $currentAuthMethodID;
			$currentTripletCombinationInfo{$CHARGING_OPTION_ID_KEY} = $currentChargingOptionID;

			#TODO - Does it need to be currentTripletCombinationInfo (Or just ..TripletInfo)
			my $tripletWeight = $this
			    ->computeTripletWeight($currentIDProviderWeight,
						   $currentAuthMethodWeight, $currentChargingOptionWeight);
			$currentTripletCombinationInfo{$WEIGHT_KEY} = $tripletWeight;

			my @allInfoDesignatorElements =
			    (@$authMethodInfoDesignatorElements, @$chargingOptionInfoDesignatorElements);

			if ($PRINT_INFORMATION)
			{
			    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
				  "Combination has weight = \'" . $currentTripletCombinationInfo{$WEIGHT_KEY} .
				  "\' and number of info designators = \'" . scalar(@allInfoDesignatorElements) .
				  "\'.\n");
			}

			$currentTripletCombinationInfo{$INFO_DESIGNATOR_ELEMENTS_KEY} =
			    \@allInfoDesignatorElements;


			my $candidateTripletInfoIndex = 0;
			for ($candidateTripletInfoIndex = 0;
			     $candidateTripletInfoIndex < scalar(@tripletCombinationInfo);
			     $candidateTripletInfoIndex ++)
			{
			    my $candidateTripletInfo = $tripletCombinationInfo[$candidateTripletInfoIndex];
			    my $candidateTripletWeight = $candidateTripletInfo->{$WEIGHT_KEY};

			    if ($PRINT_INFORMATION)
			    {
				print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
				      "Placing triplet in array; Comparing to triplet with weight = \'" .
				      $candidateTripletWeight . "\'.\n");
			    }

			    #TODO - Could change this comparison op to >= to improve efficiency
			    if ($tripletWeight > $candidateTripletWeight)
			    {
				last;
			    }
			}

			if ($PRINT_INFORMATION)
			{
			    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
				  "Placing info at index = \'" . $candidateTripletInfoIndex . "\' in array.\n");
			}

			#TODO - Note about the efficiency of splice
			splice(@tripletCombinationInfo, $candidateTripletInfoIndex, 0, \%currentTripletCombinationInfo);
		    }
		}
	    }
	}
    }

    return \@tripletCombinationInfo;
}

#TODO - Talk about the interaction between this and getGroupMemberWeight (Especially about the null "pseudo-group"
sub getPolicyPreferenceGroup
{
    my $PRINT_LOCAL_PREFIX = 'getPolicyPreferenceGroup() - ';

    my $this = shift;
    my ($policyPreferencesParseTree, $desiredGroupType) = @_;

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Looking for PreferenceGroup with type = \'" . $desiredGroupType . "\'.\n");
    }

    my @preferenceGroupElements = $policyPreferencesParseTree->getElementsByTagName($PREFERENCE_GROUP_ELEMENT_NAME);

    my $desiredGroupElement;
    foreach my $currentGroupElement (@preferenceGroupElements)
    {
	my $currentGroupType = $currentGroupElement->getAttribute($TYPE_ATTRIBUTE_NAME);

	if ($PRINT_INFORMATION)
	{
	    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		  "Looking at PreferenceGroup with type = \'" . $currentGroupType . "\'.\n");
	}

	if ($currentGroupType eq $desiredGroupType)
	{
	    if ($PRINT_INFORMATION)
	    {
		print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		      "Found a match.\n");
	    }

	    $desiredGroupElement = $currentGroupElement;

	    last;
	}
    }

    return $desiredGroupElement;
}

sub getPolicyPreferenceEquation
{
    my $PRINT_LOCAL_PREFIX = 'getPolicyPreferenceEquation() - ';

    my $this = shift;
    my ($policyPreferencesParseTree, $desiredEquationType) = @_;

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Looking for equation with type = \'" . $desiredEquationType . "\'.\n");
    }

    my $desiredEquation = $NULL_EQUATION_NAME;

    my @weightingEquationElements = $policyPreferencesParseTree
	->getElementsByTagName($WEIGHTING_EQUATION_ELEMENT_NAME);

    foreach my $currentWeightingEquationElement (@weightingEquationElements)
    {
	my $currentWeightingEquationType = $currentWeightingEquationElement
	    ->getAttribute($TYPE_ATTRIBUTE_NAME);
	
	if ($PRINT_INFORMATION)
	{
	    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		  "Looking at equation with type = \'" . $currentWeightingEquationType . "\'.\n");
	}
	
	if ($currentWeightingEquationType eq $desiredEquationType)
	{
	    $desiredEquation = $currentWeightingEquationElement
		->getAttribute($EQUATION_ATTRIBUTE_NAME);
	    
	    if ($PRINT_INFORMATION)
	    {
		print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		      "Found a match.\n");
	    }
	    
	    last;
	}
    }
    
    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Returning equation = \'" . $desiredEquation . "\'.\n");
    }
    
    return $desiredEquation;
}

sub evaluateEquation
{
    my $PRINT_LOCAL_PREFIX = 'evaluateEquation() - ';

    my $this = shift;
    my ($chargingOptionWeightingEquation, $currentChargingOptionInfo) = @_;

    my $calculatedWeight = $ZERO_WEIGHT;

    if ($chargingOptionWeightingEquation eq $LINEAR_EQUATION_NAME)
    {
	if ($PRINT_INFORMATION)
	{
	    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		  "Equation is linear.\n");
	}
	
	my $timeUnit = $currentChargingOptionInfo->{$TIME_UNIT_KEY};
	my $unitPrice = $currentChargingOptionInfo->{$UNIT_PRICE_KEY};

	if ($timeUnit && $unitPrice)
	{
	    if ($PRINT_INFORMATION)
	    {
		print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		      "There is enough information to evaluate the equation.\n");
	    }

	    # TODO - Explain this
	    $calculatedWeight = $timeUnit / $unitPrice;
	}
	else
	{
	    if ($PRINT_INFORMATION)
	    {
		print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		      "There is not enough information to evaluate the equation.\n");
	    }
	}
    }
    else
    {
	# This will fall through to simply returning a 0 weight (In the same way getGroupMemberWeight
	#   will return 0 on an unkown (null) group).
	if ($PRINT_INFORMATION)
	{
	    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		  "The equation is of unknown form = \'" . $chargingOptionWeightingEquation . "\'.\n");
	}
    }

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Returning a weight = \'" . $calculatedWeight . "\'.\n");
    }

    return $calculatedWeight;
    #TODO - Need to deal with possibly null values (I think that results in a 0 value for now).
}

sub computeTripletWeight
{
    my $this = shift;
    my ($IDProviderWeight, $authMethodWeight, $chargingOptionWeight) = @_;

    my $totalWeight = $IDProviderWeight * $authMethodWeight * $chargingOptionWeight;

    return $totalWeight;
}

sub getInfoDesignatorNamesFromElements
{
    my $this = shift;
    my ($infoDesignatorElements, $desiredAttributeName) = @_;

    my @infoDesignatorNames;

    foreach my $currentInfoDesignatorElement (@$infoDesignatorElements)
    {
	my $currentInfoDesignatorName = $currentInfoDesignatorElement->getAttribute($desiredAttributeName);
	
	push(@infoDesignatorNames, $currentInfoDesignatorName);
    }

    return \@infoDesignatorNames;
}

sub getPreferenceGroupMemberWeight
{
    my $PRINT_LOCAL_PREFIX = 'getPreferenceGroupMemberWeight() - ';

    my $this = shift;
    my ($preferenceGroupElement, $desiredGroupMemberName) = @_;

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Looking for GroupMember with name = \'" . $desiredGroupMemberName . "\'.\n");
    }

    if (!($preferenceGroupElement))
    {
	if ($PRINT_INFORMATION)
	{
	    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		  "Got null preference group; Returning 0 weight.\n");
	}

	return $ZERO_WEIGHT;
    }

    #TODO - Here we default to 0 if the group member is not found
    my $desiredGroupMemberWeight = $ZERO_WEIGHT;

    my @preferenceGroupMemberElements = $preferenceGroupElement
	->getChildrenByTagName($PREFERENCE_GROUP_MEMBER_ELEMENT_NAME);

    foreach my $currentGroupMemberElement (@preferenceGroupMemberElements)
    {
	my $currentGroupMemberName = $currentGroupMemberElement->getAttribute($GROUP_MEMBER_NAME_ATTRIBUTE_NAME);

	if ($PRINT_INFORMATION)
	{
	    print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		  "Looking at GroupMember with name \'" . $currentGroupMemberName . "\'.\n");
	}

	if ($currentGroupMemberName eq $desiredGroupMemberName)
	{
	    if ($PRINT_INFORMATION)
	    {
		print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
		      "Found a match.\n");
	    }

	    $desiredGroupMemberWeight = $currentGroupMemberElement
		->getAttribute($GROUP_MEMBER_WEIGHT_ATTRIBUTE_NAME);

	    last;
	}
    }

    if ($PRINT_INFORMATION)
    {
	print(STDERR $PRINT_INFORMATION_PREFIX . $PRINT_LOCAL_PREFIX .
	      "Returning weight = \'" . $desiredGroupMemberWeight . "\'.\n");
    }
    
    return $desiredGroupMemberWeight;
}

sub printAuthMethodInfo
{
    my $this = shift;
    my ($authMethodInfo) = @_;

    print("===== AUTH METHOD INFO =====\n");

    foreach my $currentAuthMethodID (keys(%$authMethodInfo))
    {
	print("  Auth method = \'" . $currentAuthMethodID . "\':\n");
	my $currentAuthMethodInfo = $authMethodInfo->{$currentAuthMethodID};

	print("    Weight = \'" . $currentAuthMethodInfo->{$WEIGHT_KEY} . "\'.\n");
	
	my $infoDesignatorElementsReference = $currentAuthMethodInfo->{$INFO_DESIGNATOR_ELEMENTS_KEY};
	my @currentInfoDesignatorElements = @$infoDesignatorElementsReference;
	print("    Number of info designators = \'" . scalar(@currentInfoDesignatorElements) . "\'.\n");
    }
}

sub printChargingOptionInfo
{
    my $this = shift;
    my ($chargingOptionInfo) = @_;

    print("===== CHARGING OPTION INFO =====\n");
    
    foreach my $currentChargingOptionID (keys(%$chargingOptionInfo))
    {
	print("  Charging option = \'" . $currentChargingOptionID . "\':\n");
	my $currentChargingOptionInfo = $chargingOptionInfo->{$currentChargingOptionID};

	my $timeUnit = $currentChargingOptionInfo->{$TIME_UNIT_KEY};
	if ($timeUnit)
	{
	    print("    Time unit = \'" . $timeUnit . "\'.\n");
	}
	else
	{
	    print("    Time unit is null.\n");
	}

	my $unitPrice = $currentChargingOptionInfo->{$UNIT_PRICE_KEY};
	if ($unitPrice)
	{
	    print("    Unit price = \'" . $unitPrice . "\'.\n");
	}
	else
	{
	    print("    Unit price is null.\n");
	}

	print("    Weight = \'" . $currentChargingOptionInfo->{$WEIGHT_KEY} . "\'.\n");

	my $infoDesignatorElementsReference = $currentChargingOptionInfo->{$INFO_DESIGNATOR_ELEMENTS_KEY};
	my @infoDesignatorElements = @$infoDesignatorElementsReference;
	print("    Number of info designators = \'" . scalar(@infoDesignatorElements) . "\'.\n");
    }
}

sub printIDProviderInfo
{
    my $this = shift;
    my ($IDProviderInfo) = @_;

    print("===== ID PROVIDER INFO =====\n");

    foreach my $currentIDProviderID (keys(%$IDProviderInfo))
    {
	my $currentIDProviderInfo = $IDProviderInfo->{$currentIDProviderID};
	
	print("  ID Provider = \'" . $currentIDProviderID . "\':\n");
	print("    Weight = \'" . $currentIDProviderInfo->{$WEIGHT_KEY} . "\'.\n");
    }
}

sub printTripletCombinationInfo
{
    my $this = shift;
    my ($tripletCombinationInfo) = @_;

    print("===== TRIPLET INFO =====\n");

    for (my $currentTripletInfoIndex = 0;
	 $currentTripletInfoIndex < scalar(@$tripletCombinationInfo);
	 $currentTripletInfoIndex ++)
    {
	my $currentTripletInfo = $tripletCombinationInfo->[$currentTripletInfoIndex];
	my $IDProviderName = $currentTripletInfo->{$ID_PROVIDER_NAME_KEY};
	my $authMethodID = $currentTripletInfo->{$AUTH_METHOD_ID_KEY};
	my $chargingOptionID = $currentTripletInfo->{$CHARGING_OPTION_ID_KEY};
	print("  (" . $IDProviderName . ", " . $authMethodID . ", " . $chargingOptionID . "):\n");

	my $tripletWeight = $currentTripletInfo->{$WEIGHT_KEY};
	print("    Weight = \'" . $tripletWeight . "\'.\n");

	my $infoDesignatorElements = $currentTripletInfo->{$INFO_DESIGNATOR_ELEMENTS_KEY};
	print("    Number of info designators = \'" . scalar(@$infoDesignatorElements) . "\'.\n");
    }
}

return $TRUE;
