#!/usr/bin/perl -w
#Wireless LAN Inter-System Roaming Helper (wrh.pl)
#Authors: Yasuhiko MATSUNAGA, Ana SANZ MERINO
#E-mail: asanz@eecs.berkeley.edu
#Revision History
#ver.0.1 03/25/03
#ver.0.2 04/04/03 - Added DHCP/Static IP handling at handover 
#                   (currently only ISC DHCP client (dhclient) is supported)
#ver.0.3 06/25/03 - Added web-based login with credentials in Policy Engine
#ver.0.4 04/12/04 - Added 802.1x authentication
#                   Repaired bugs in DHCP handling and wireless params setting
#ver.0.5 01/21/05 - Added compound L2-Web_based authentication
#                   Extended configuration options
#                   Added connection attempt to unconfigured networks
#                   Note: Although the roaming client supports different 
#                   authentication options, such as static wireless encryption,
#                   only the case of compound L2-web based authentication has
#                   been fully implemented and tested

BEGIN
{
    my $MODULE_INCLUDE_PATH = './';

    @INC = (@INC, $MODULE_INCLUDE_PATH);
}

use Time::HiRes qw(usleep ualarm gettimeofday tv_interval);
use strict;
use Tk;
use Tk::BrowseEntry;
use IO;
use XML::Writer;                           # Able to write XML files
use LWP::UserAgent;                        # Web Client
use HTTP::Request;
use HTTP::Request::Common qw(POST);        # POST method support
use HTTP::Response;
use HTTP::Cookies;
use SOAP::Lite;                            # Used in interaction with ANP Server
use XML::Parser;
use ANPClient;

#Define Global Variables
my $lib_path = "/usr/local/lib:/usr/local/ssl/lib";
my $xsupplicant_conf = "/etc/1x/1x.conf";
# Program for extracting L2 authn info
my $l2_info_extracter = "./get_l2_info.pl";
my $ROAMING_CLIENT_CONFIG_FILE_PATH = 'config/roamingClient.conf';
my $linux = 1;                  #for testing
my $mw;                         #Main window
my $sw1;                        #Per-System configuration sub-window
my $sw2;                        #General configuration sub-window
my $sw3;                        #Manual IP configuration sub-window
my (@list1,@list2);             #Main window listbox (configured/unconfigured)
my $config_file_name;           #Configuration file name
my @hidden_list = ();           #Hidden SSID list
my %ifc_config_list;            #Configuration information per WLAN interface
my $interval = 2000;            #Periodic Update Interval (default 2 sec) 
my $selected_ssid = '';         #Selected SSID in listbox
my $selected_index = -1;        #Index of selected configured system 
my $server_comm_mode = "disabled";      #Management Server Communication
my $pending_update_id;          #One-time refresh event ID after autoconnect
my $periodic_update_id;         #Periodic refresh event ID      
my @wlan_ifc_list;              #WLAN Interface List
my $wlan_ifc;                   #Selected WLAN Interface 
my %current_ap;                 #Current Wireless LAN system information
my (@sys_list1, @sys_list2);    #Configured/Unconfigured Wireless LAN system information
my $conf_list_ref;              #Reference to the currently-selected configuration information
my @ap_list;                    #Scanning result
my $mac_addr;
my $l2_key_digest;
my $anpClient = ANPClient->new();

################################################################################
### Main routine
################################################################################

$ENV{'LD_LIBRARY_PATH'} = $lib_path;

#Get Configuration File Name
$config_file_name = &GetConfigFileName(@ARGV);
 
#Read Configuration Data from File
if ($config_file_name ne '') {
    &ReadConfigFile($config_file_name, \$interval, \%ifc_config_list,
                    \@hidden_list);
}

#Get WLAN Interface List
@wlan_ifc_list = &GetIfcList();

#Select WLAN Interface
$wlan_ifc = $wlan_ifc_list[0];
#print("wlan_ifc_list = @wlan_ifc_list, wlan_ifc = $wlan_ifc\n");

#Get Configuration Data for Selected WLAN Interface
$conf_list_ref = $ifc_config_list{$wlan_ifc};

#Remove current essid, in case it was configured in the card ifcfg configuration
# file but we do not really have connectivity to that network
`/sbin/iwconfig $wlan_ifc essid ""`;

#Get Current AP Info
&GetCurentApInfo($wlan_ifc, \%current_ap);

#Get Other AP List
&GetApList($wlan_ifc, \@hidden_list, \@ap_list);

#Prepare Data for Display
&UpdateSysInfo($conf_list_ref, \%current_ap, \@ap_list, \@sys_list1, \@sys_list2);

#Initialize Main Window 
$mw = new MainWindow;
$mw->title("WLAN Inter-System Roaming Helper");
$mw->optionAdd('*font' => 'Arial 14');

#Automatic Connect
&AutomaticConnect(\@sys_list1, \@sys_list2);

#Display WLAN Interface List 
&DisplayWlanIfList($mw, \@wlan_ifc_list, \$wlan_ifc);

#Display Available/Configured WLAN Systems
&DisplayWlanSystems($mw, \@sys_list1, \@sys_list2, \@list1, \@list2);

#Display Buttons
&DisplayButtons($mw, $sw1, $sw2);

#Define Periodic Update
if ($interval > 0) {
    $periodic_update_id = $mw->repeat($interval, \&PeriodicScan);
}

#Wait for Events
MainLoop();


################################################################################
### Callback Functions (main window)
################################################################################

################################################################################
### ChangeIfc ($selected_wlan_ifc_ref)
### Change wireless LAN interface and re-initialize 
### Input: $selected_wlan_ifc_ref - newly selected wireless LAN interface name
################################################################################
sub ChangeIfc {
    my $selected_wlan_ifc_ref = $_[0];
    my $i;
    print("Interface Changed to $$selected_wlan_ifc_ref\n");
    #Reselect WLAN Interface
    $wlan_ifc = $$selected_wlan_ifc_ref;
    
    #Get Configuration Data for Re-selected WLAN Interface
    $conf_list_ref = $ifc_config_list{$wlan_ifc};
    
    #Get Current AP Info
    &GetCurentApInfo($wlan_ifc, \%current_ap);
    
    #Get Other AP List
    &GetApList($wlan_ifc, \@hidden_list, \@ap_list);
    
    #Prepare Data for Display
    &UpdateSysInfo($conf_list_ref, \%current_ap, \@ap_list, \@sys_list1, \@sys_list2);
    
    #Display Available/Configured WLAN Systems
    &ReDisplayWlanSystems(\@sys_list1, \@sys_list2, \@list1, \@list2);
    
    #Automatic Re-connect
    &AutomaticConnect(\@sys_list1, \@sys_list2);
}

################################################################################
### GetSelectedSsid
### Get selected wireless system ID from listbox
################################################################################
sub GetSelectedSsid {
    my @index_list;
    @index_list = $list1[0]->curselection;
    if (@index_list) {
        #Select only the first system
        $selected_ssid = $sys_list1[$index_list[0]]{SSID};
        $selected_index = $index_list[0];
        print ("ssid = $selected_ssid\n");
        return;
    }
    @index_list = $list2[0]->curselection;
    if (@index_list) {
        #Select only the first system
        $selected_ssid = $sys_list2[$index_list[0]]{SSID};
        $selected_index = -1;
        print ("ssid = $selected_ssid\n");
        return;
    }
}

################################################################################
### MultiScrollSet1
### Routine for scrolling listboxes for configured systems altogether
################################################################################
sub MultiScrollSet1 {
    print "@_";
    my $widget;
    foreach $widget (@list1) {
        $widget->yview(@_);
    }
}

################################################################################
### MultiScrollSet2
### Routine for scrolling listboxes for unconfigured systems altogether
################################################################################
sub MultiScrollSet2 {
    my $widget;
    foreach $widget (@list2) {
        $widget->yview(@_);
    }
}


################################################################################
### PriorityUp
### Give higher priority to the selected system, update display and handover
################################################################################
sub PriorityUp {
    my $tmp;
    if ($selected_index < 0) {
        $mw->messageBox(-title => 'Warning',
                        -icon => 'warning',
                        -type => 'ok',
                        -message => "Invalid priority change for unconfigured system $selected_ssid");
        return;
    } elsif ($selected_index == 0) {
        $mw->messageBox(-title => 'Warning',
                        -icon => 'warning',
                        -type => 'ok',
                        -message => "Invalid priority change for the highest priority system $selected_ssid");
        return;
    }
    #Swap system configuration array
    $tmp = $conf_list_ref->[$selected_index-1];
    $conf_list_ref->[$selected_index-1] = $conf_list_ref->[$selected_index];
    $conf_list_ref->[$selected_index] = $tmp;   
    
    #Swap configured system list
    $tmp = $sys_list1[$selected_index-1];
    $sys_list1[$selected_index-1] = $sys_list1[$selected_index];
    $sys_list1[$selected_index] = $tmp;
    
    $selected_index--;
    &UpdateDisplay();
    &WriteConfigFile($config_file_name, \$interval, \%ifc_config_list);
    &AutomaticConnect(\@sys_list1, \@sys_list2);
}

################################################################################
### PriorityDown
### Give lower priority to the selected system, update display and handover
################################################################################
sub PriorityDown {
    my $tmp;
    if ($selected_index < 0) {
        $mw->messageBox(-title => 'Warning',
                        -icon => 'warning',
                        -type => 'ok',
                        -message => "Invalid priority change for unconfigured system $selected_ssid");
        return;
    } elsif ($selected_index == $#sys_list1) {
        $mw->messageBox(-title => 'Warning',
                        -icon => 'warning',
                        -type => 'ok',
                        -message => "Invalid priority change for the lowest priority system $selected_ssid");
        return;
    }
    #Swap system configuration array
    $tmp = $conf_list_ref->[$selected_index+1];
    $conf_list_ref->[$selected_index+1] = $conf_list_ref->[$selected_index];
    $conf_list_ref->[$selected_index] = $tmp;   

    #Swap configured system list
    $tmp = $sys_list1[$selected_index+1];
    $sys_list1[$selected_index+1] = $sys_list1[$selected_index];
    $sys_list1[$selected_index] = $tmp;
    
    $selected_index++;
    &UpdateDisplay();
    &WriteConfigFile($config_file_name, \$interval, \%ifc_config_list);
    &AutomaticConnect(\@sys_list1, \@sys_list2);
}

################################################################################
### UpdateDisplay
### Scan current/neighbor cells and update listbox information
### Global variables used: $wlan_ifc, %current_ap, @hidden_list, @ap_list,
###                        @sys_list1, @sys_list2, @list1, @list2,
###                        $selected_ssid 
################################################################################
sub UpdateDisplay {
    my $i;
    my $sysinfo_ref;
    my $size;

    #Get Current AP Info
    &GetCurentApInfo($wlan_ifc, \%current_ap);
    #Get Other AP List
    &GetApList($wlan_ifc, \@hidden_list, \@ap_list);
    &UpdateSysInfo($conf_list_ref, \%current_ap, \@ap_list, \@sys_list1, \@sys_list2);
    #Listbox Update
    &ReDisplayWlanSystems(\@sys_list1, \@sys_list2, \@list1, \@list2, \$selected_ssid);
}


################################################################################
### Callback Functions (sub window)
################################################################################

################################################################################
### OpenPersysConfigWindow
### Open Per-system configuration window
### Callback: ChangeSsid - on selecting another SSID in System ID browseentry
###           UpdatePersysConfigInfo - on pressing 'Add/Change' button
###           DeletePersysConfigInfo - on pressing 'Delete' button
###           CancelPersysConfigInfo - on pressing 'Cancel' button
### Global variables used: @sys_list1, $selected_ssid, 
################################################################################
sub OpenPersysConfigWindow {
    my ($frame1, $frame2);
    my $hash_key;
    my @tmp_ssid_list = ();
    my %tmp_sysinfo;
    my $system_entry;
    my $sysinfo_ref;
    my $already_open;
    my ($add_button, $del_button, $ok_button, $ca_button);
    my $ip_mode;
    my $i;
    my @labels = ();
    my @f1_label;
    my @encrypt_mode_list;
    my @ip_mode_list;
    my @encrypt_mode_value;
    my @rad;
    my @chk;
    my @ent;
    my @but;
    my $tmp_encrypt_mode;
    
    #Do not open duplicate sub-window if per-system 
    if (Exists($sw1)) {
        return; 
    }
    
    #Default Settings
    $tmp_sysinfo{SSID} = "";
    $tmp_sysinfo{HIDDEN} = "no";
    $tmp_sysinfo{ENCRYPT} = "off";
    $tmp_sysinfo{IPMODE}  = "dhcp";

    #Prepare Configured/Detected System List
    foreach $sysinfo_ref (@sys_list1) {
        #Make configured system list for browsing
        push(@tmp_ssid_list, $sysinfo_ref->{SSID});
        #Copy temporary field for configuation change 
        if ($sysinfo_ref->{SSID} eq $selected_ssid) {
            #copy sysinfo hash into temporary fields
            foreach $hash_key (keys %{$sysinfo_ref}) {
                $tmp_sysinfo{$hash_key} = $sysinfo_ref->{$hash_key};
                #print("OpenPersysConfigWindow(1): hash_key = $hash_key, sysinfo = $tmp_sysinfo{$hash_key}\n");
            }
        }
    }
    foreach $sysinfo_ref (@sys_list2) {
        if ($sysinfo_ref->{SSID} eq "hidden") {
            next;
        }
        push(@tmp_ssid_list, $sysinfo_ref->{SSID});
        if ($sysinfo_ref->{SSID} eq $selected_ssid) {
            #copy sysinfo hash into temporary fields
            foreach $hash_key (keys %{$sysinfo_ref}) {
                $tmp_sysinfo{$hash_key} = $sysinfo_ref->{$hash_key};
                #print("OpenPersysConfigWindow(2): hash_key = $hash_key, sysinfo = $tmp_sysinfo{$hash_key}\n");
            }
        }
    }
    push(@tmp_ssid_list, "(new)");
    
    print("ssid = $tmp_sysinfo{SSID}\n");
    print("tmp_ssid_list = @tmp_ssid_list\n");
    print("\n");

    #Open Sub-window
    $sw1 = $mw->Toplevel();
    $sw1 -> title("Per-System Configuration");       
    
    @f1_label = ("Per-System Configuration", "\tSystem ID", "\tEncryption",
                 "\tKey", "\tCertificate Path","","\tIP Settings");
    #Per-System Configuration
    $frame1 = $sw1->Frame(-borderwidth =>2, -relief => 'raised')
        ->pack(-side =>'top', -padx =>3, -pady =>3);
    for ($i = 0; $i < 7; $i++) {
        $frame1 -> Label(-text => $f1_label[$i],-anchor=> 'nw')
	    -> grid(-row => $i, -column  => 0, -sticky  => 'news',);
    }

    #System ID BrowseEntry
    $frame1 -> BrowseEntry(-variable=> \$tmp_sysinfo{SSID},
			   -choices => \@tmp_ssid_list,
			   -browsecmd => [\&ChangeSsid, \%tmp_sysinfo, \@chk, \@rad, \@ent])
        -> grid(-row => 1, -column  => 1, -columnspan => 2, -sticky  => 'news',);
    #Hidden system ID CheckButton
    $chk[0] = $frame1 -> Checkbutton(-text => "(hidden ID)  ", 
				     -variable => \$tmp_sysinfo{HIDDEN}, 
				     -onvalue => "yes",
				     -offvalue => "no")
        -> grid(-row => 1, -column  => 3, -sticky  => 'news');
    if ($tmp_sysinfo{HIDDEN} eq 'yes') {
        $chk[0]->select;
    }
    
    @encrypt_mode_list = ("off", "static", "auto(802.1x)");
    @encrypt_mode_value = ("off", "static", "auto");
    $tmp_encrypt_mode = $tmp_sysinfo{ENCRYPT};
    $tmp_sysinfo{ENCRYPT} = "";     #for deselecting radiobutton
    for ($i = 0; $i < 3; $i++) {
        $rad[$i] = $frame1 -> Radiobutton(-text => $encrypt_mode_list[$i], 
					  -variable => \$tmp_sysinfo{ENCRYPT}, 
					  -value => $encrypt_mode_value[$i])
	    -> grid(-row => 2, -column  => $i+1, -sticky  =>'news',);
        if ($tmp_encrypt_mode eq $encrypt_mode_value[$i]) {
            $rad[$i] -> select();
        }
    }
    $tmp_sysinfo{ENCRYPT} = $tmp_encrypt_mode;
    
    #Encryption key entry
    $ent[0] = $frame1 -> Entry(-width => 20, -show => '*')
        -> grid(-row => 3, -column  => 1, -columnspan => 3, -sticky => 'news');
    if ($tmp_sysinfo{KEY}) {
        $ent[0] -> insert('end', $tmp_sysinfo{KEY});
    }
    $ent[0]->bind('<Key>', sub {$tmp_sysinfo{KEY} = $ent[0]->get});

    #Certificate path entry
    $ent[1] = $frame1 -> Entry(-width => 20)
        -> grid(-row => 4, -column  => 1, -columnspan => 3, -sticky => 'news');
    if (defined $tmp_sysinfo{CERTPATH}) {
        $ent[1] -> insert('end', $tmp_sysinfo{CERTPATH});
    }

    $but[0] = $frame1 -> Button(-text => "Browse Certificate", 
                                -width => 15, 
                                -command => [\&BrowseCertificate, \$ent[1], \$tmp_sysinfo{CERTPATH}] )
        -> grid(-row => 5, -column  => 1, -sticky => 'e');

    $rad[3] = $frame1 -> Radiobutton(-text => "dhcp", 
                                     -variable => \$tmp_sysinfo{IPMODE}, 
                                     -value => "dhcp")
        -> grid(-row => 6, -column  => 1, -sticky  => 'news');
    $rad[4] = $frame1 -> Radiobutton(-text => "static", 
                                     -variable => \$tmp_sysinfo{IPMODE}, 
                                     -value => "static")
        -> grid(-row => 6, -column  => 2, -sticky  => 'news');
    $but[1] = $frame1 -> Button( -text => "Manual IP Config", -width => 15,
                                 -command => [\&OpenManualIpConfigWindow, \%tmp_sysinfo])
        -> grid(-row => 6, -column  => 3, -sticky => 'e');

    #OK/Cancel Buttons
    $frame2 = $sw1->Frame()
        ->pack(-side =>'right', -padx =>3, -pady =>3);
    
    $but[2] = $frame2 -> Button(-text => "Add/Change", 
                                -height => 1, -width =>12,
                                -command => [\&UpdatePersysConfigInfo, \%tmp_sysinfo] )
        ->pack(-side =>'left', -padx => 3, -pady => 3);
    $but[3] = $frame2 -> Button(-text => "Delete", 
                                -height => 1, -width =>12,
                                -command => [\&DeletePersysConfigInfo, \%tmp_sysinfo] )
        ->pack(-side =>'left', -padx => 3, -pady => 3);
    $but[4] = $frame2 -> Button(-text => "Cancel", 
                                -height => 1, -width =>12,
                                -command => [\&CancelPersysConfigInfo])
        ->pack(-side =>'left', -padx => 3, -pady => 3);
}

################################################################################
### BrowseCertificate ($ent_ref, $certpath_ref)
### Browse certificate file
### Input: $ent_ref - reference to the entry box for displaying certificate path
###        $certpath_ref - reference to selected certificate path
################################################################################
sub BrowseCertificate {
    my ($ent_ref, $certpath_ref) = @_;  
    my $filename;
    $filename = $sw1->getOpenFile();
    if (defined $filename) {
        $$ent_ref->delete(0,'end');
        $$ent_ref->insert('end', $filename);
        $$certpath_ref = $filename;
    }
}

################################################################################
### OpenManualIpConfigWindow ($tmp_sysinfo_ref)
### Manual TCP/IP Configuration sub-window
### Input/Output: $tmp_sysinfo_ref - reference to temporary per-system
###                                  configuration information buffer
################################################################################
sub OpenManualIpConfigWindow {
    my $tmp_sysinfo_ref= $_[0];
    my ($frame1, $frame2);
    my (@ip_addr_entry, @ip_mask_entry, @ip_dgwy_entry, @dns_pri_entry, @dns_sec_entry);
    my ($i, $j);
    my ($conf_ref, $ssid, @words);

    #Open Sub-window
    if (Exists($sw3)) {
	return; 
    }
    $sw3 = $sw1->Toplevel();
    $sw3 -> title("Manual IP Configuration");
    
    #General Configuration
    $frame1 = $sw3->Frame(-borderwidth =>2, -relief => 'raised')
        ->pack(-side =>'top', -padx =>3, -pady =>3);
    $frame1 -> Label(-text => 'Manual IP Configuration ', -anchor=> 'nw')
        -> grid(-row => 0, -column  => 0, -sticky  => 'news');
    $frame1 -> Label(-text => "\tIP Address", -anchor=> 'w')
        -> grid(-row => 1, -column  => 0, -sticky  => 'news');
    $frame1 -> Label(-text => "\tIP Netmask", -anchor=> 'w')
        -> grid(-row => 2, -column  => 0, -columnspan =>2, -sticky  => 'news');
    $frame1 -> Label(-text => "\tDefault Gateway", -anchor=> 'w')
        -> grid(-row => 3, -column  => 0, -columnspan =>2, -sticky  => 'news');
    $frame1 -> Label(-text => "\tDNS Primary", -anchor=> 'w')
        -> grid(-row => 4, -column  => 0, -sticky  => 'news');
    $frame1 -> Label(-text => "\tDNS Secondary", -anchor=> 'w')
        -> grid(-row => 5, -column  => 0, -sticky  => 'news');
    
    for ($i = 0; $i < 4; $i++) {
        $ip_addr_entry[$i] = $frame1 -> Entry(-width => 5)
            -> grid(-row => 1, -column  => 2*$i+1, -sticky  => 'news');
        $ip_mask_entry[$i] = $frame1 -> Entry(-width => 5)
            -> grid(-row => 2, -column  => 2*$i+1, -sticky  => 'news');
        $ip_dgwy_entry[$i] = $frame1 -> Entry(-width => 5)
            -> grid(-row => 3, -column  => 2*$i+1, -sticky  => 'news');
        $dns_pri_entry[$i] = $frame1 -> Entry(-width => 5)
            -> grid(-row => 4, -column  => 2*$i+1, -sticky  => 'news');
        $dns_sec_entry[$i] = $frame1 -> Entry(-width => 5)
            -> grid(-row => 5, -column  => 2*$i+1, -sticky  => 'news');
    }

    for ($i = 0; $i < 3; $i++) {
        for ($j = 1; $j < 6; $j++) {
            $frame1 -> Label(-text => ".", -anchor=> 'w')
                -> grid(-row => $j, -column  => 2*$i+2, -sticky  => 'news');    
        }
    }

    #Enter Current values
    #Find existing configuration entry and enter it

    foreach $conf_ref (@{$conf_list_ref}) {
        $ssid = $conf_ref->{SSID};
        if ($conf_ref->{SSID} ne $tmp_sysinfo_ref->{SSID}) {
            next;
        } else {
            #found existing entry;
            if (defined $conf_ref->{IPADDR}) {
                @words = split(/\./, $conf_ref->{IPADDR});
                for ($i = 0; $i < 4; $i++) {
                    $ip_addr_entry[$i]->insert('end', $words[$i]);
                }
            }
            if (defined $conf_ref->{IPMASK}) {
                @words = split(/\./, $conf_ref->{IPMASK});
                for ($i = 0; $i < 4; $i++) {
                    $ip_mask_entry[$i]->insert('end', $words[$i]);
                }
            }
            if (defined $conf_ref->{IPDGWY}) {
                @words = split(/\./, $conf_ref->{IPDGWY});
                for ($i = 0; $i < 4; $i++) {
                    $ip_dgwy_entry[$i]->insert('end', $words[$i]);
                }
            }
            if (defined $conf_ref->{DNSPRI}) {
                @words = split(/\./, $conf_ref->{DNSPRI});
                for ($i = 0; $i < 4; $i++) {
                    $dns_pri_entry[$i]->insert('end', $words[$i]);
                }
            }
            if (defined $conf_ref->{DNSSEC}) {
                @words = split(/\./, $conf_ref->{DNSSEC});
                for ($i = 0; $i < 4; $i++) {
                    $dns_sec_entry[$i]->insert('end', $words[$i]);
                }
            }
            last;
        }
    }

    #OK/Cancel Buttons
    $frame2 = $sw3->Frame()
        ->pack(-side =>'right', -padx =>3, -pady =>3);
    $frame2 -> Button(-text => "Ok", -height => 1, -width =>8,
		      -command => [\&SaveManualIpInfo, 
				   $tmp_sysinfo_ref,
				   \@ip_addr_entry,
				   \@ip_mask_entry,
				   \@ip_dgwy_entry,
				   \@dns_pri_entry,
				   \@dns_sec_entry])
        ->pack(-side =>'left', -padx =>3, -pady =>3);
    $frame2 -> Button(-text => "Cancel", -height => 1, -width =>8,
		      -command => \&CancelManualIpInfo)
        ->pack(-side =>'left', -padx =>3, -pady =>3);
}

################################################################################
### CheckRange ($entry_ref, $adr_ref, $type_ref)
### Retrieve manual IP configuration information and check validity
### Input: $entry_ref - reference to array of IP address entry box
###        $type_ref - reference to information type string (IP, MASK, DNS,etc.)
### Input/Output: $adr_ref - reference to array of IP address/netmask
################################################################################
sub CheckRange {
    my ($entry_ref, $adr_ref, $type_ref) = @_;
    my $i;
    my $flag = 0;

    print "CheckRange: $$type_ref = ";
    for ($i = 0; $i < 4; $i++) {
        $adr_ref->[$i] = $entry_ref->[$i]->get;
        unless ($flag) {
            if ($adr_ref->[$i] ne "") {
                $flag = 1;
            }
        }
        if ($adr_ref->[$i] ne "") {
            if ( ($adr_ref->[$i] < 0) || ($adr_ref->[$i] > 255) ) {
                $sw3->messageBox(-title => 'Warning',
				 -icon => 'warning',
				 -type => 'ok',
				 -message => "Invalid $$type_ref");
                $entry_ref->[$i]->delete(0,'end');
                return -1;
            }
            print "$adr_ref->[$i]."
	    }
    }
    print "\n";
    return $flag;
}

################################################################################
### SaveManualIpInfo ($tmp_sysinfo_ref, $ip_addr_entry_ref, $ip_mask_entry_ref, 
###                   $ip_dgwy_entry_ref, $dns_pri_entry_ref,
###                   $dns_sec_entry_ref)
### Save Manual IP configuration change and close the window
### Input/Ouput - $tmp_sysinfo_ref:  reference to the temporary per-system
###                                  configuration information
###               $ip_addr_entry_ref: reference to array of IPv4 address entries
###               $ip_mask_entry_ref: reference to array of IPv4 netmask entries
###               $ip_dgwy_entry_ref: reference to array of default gateway
###                                   entries
###               $dns_pri_entry_ref: reference to array of Primary DNS entries
###               $dns_sec_entry_ref: reference to array of Secondary DNS
###                                   entries
################################################################################
sub SaveManualIpInfo {
    my ($tmp_sysinfo_ref, $ip_addr_entry_ref, $ip_mask_entry_ref,
        $ip_dgwy_entry_ref, $dns_pri_entry_ref, $dns_sec_entry_ref) = @_;
    my $type;
    my $i;
    my ($flag_ipaddr, $flag_ipmask, $flag_ipdgwy);
    my ($flag_dnspri, $flag_dnssec);
    my (@tmp_ipaddr, @tmp_ipmask, @tmp_ipdgwy);
    my (@tmp_dnspri, @tmp_dnssec); 

    $type = "IP Address";
    $flag_ipaddr = &CheckRange($ip_addr_entry_ref, \@tmp_ipaddr, \$type);
    if ($flag_ipaddr < 0) {
        return;
    } elsif ($flag_ipaddr > 0) {
        $tmp_sysinfo_ref->{IPADDR} = "$tmp_ipaddr[0].$tmp_ipaddr[1].$tmp_ipaddr[2].$tmp_ipaddr[3]";
    }
    $type = "IP Mask";
    $flag_ipmask = &CheckRange($ip_mask_entry_ref, \@tmp_ipmask, \$type);
    if ($flag_ipmask < 0) {
        return;
    } elsif ($flag_ipmask > 0) {
        $tmp_sysinfo_ref->{IPMASK} = "$tmp_ipmask[0].$tmp_ipmask[1].$tmp_ipmask[2].$tmp_ipmask[3]";
    }

    $type = "Default Gateway";
    $flag_ipdgwy = &CheckRange($ip_dgwy_entry_ref, \@tmp_ipdgwy, \$type);
    if ($flag_ipdgwy < 0) {
        return;
    } elsif ($flag_ipdgwy > 0) {
        $tmp_sysinfo_ref->{IPDGWY} = "$tmp_ipdgwy[0].$tmp_ipdgwy[1].$tmp_ipdgwy[2].$tmp_ipdgwy[3]";
    }

    $type = "Primary DNS Address";
    $flag_dnspri = &CheckRange($dns_pri_entry_ref, \@tmp_dnspri, \$type);
    if ($flag_dnspri < 0) {
        return;
    } elsif ($flag_dnspri > 0) {
        $tmp_sysinfo_ref->{DNSPRI} = "$tmp_dnspri[0].$tmp_dnspri[1].$tmp_dnspri[2].$tmp_dnspri[3]";
    }

    $type = "Secondary DNS Address";
    $flag_dnssec = &CheckRange($dns_sec_entry_ref, \@tmp_dnssec, \$type);
    if ($flag_dnssec < 0) {
        return;
    } elsif ($flag_dnssec > 0) {
        $tmp_sysinfo_ref->{DNSSEC} = "$tmp_dnssec[0].$tmp_dnssec[1].$tmp_dnssec[2].$tmp_dnssec[3]";
    }

    $sw3->withdraw;
    #Re-initialize manual IP configuration sub-window; 
    $sw3 = '';
}

################################################################################
### CancelManualIpInfo 
### Cancel Manual IP configuration change and close the window
################################################################################
sub CancelManualIpInfo {
    $sw3->withdraw;
    #Re-initialize manual IP configuration sub-window; 
    $sw3 = '';
}


################################################################################
### ChangeSsid ($tmp_sysinfo_ref, $chk_ref, $rad_ref, $ent_ref)
### Reflect selected WLAN system ID change on per-system configuration
###   sub-window
### Input: $tmp_sysinfo_ref - reference to the temporary per-system
###                           configuration information
###        $chk_ref - reference to checkbox array in per-system configuration
###                   window
###        $rad_ref - reference to radiobutton array in per-system configuration
###                   window
###        $ent_ref - reference to entry box array in per-system configuration
###                   window
################################################################################
sub ChangeSsid {
    my ($tmp_sysinfo_ref, $chk_ref, $rad_ref, $ent_ref) = @_;
    my $sysinfo_ref;
    my $hash_key;
    my $found = 0;

    print("ChangeSsid: changed to $tmp_sysinfo_ref->{SSID}\n");
    #Default settings
    $tmp_sysinfo_ref->{SSID} = "";
    $tmp_sysinfo_ref->{HIDDEN} = "no";
    $tmp_sysinfo_ref->{ENCRYPT} = "off";
    $tmp_sysinfo_ref->{KEY} = "";
    $tmp_sysinfo_ref->{IPMODE}  = "dhcp";

    #Prepare Configured/Detected System List
    foreach $sysinfo_ref (@sys_list1) {
        #Copy temporary field for configuation change 
        if ($sysinfo_ref->{SSID} eq $tmp_sysinfo_ref->{SSID}) {
            #copy sysinfo hash into temporary fields
            foreach $hash_key (keys %{$sysinfo_ref}) {
                $tmp_sysinfo_ref->{$hash_key} = $sysinfo_ref->{$hash_key};
            }
            $found = 1;
        }
    }
    unless ($found) {
        foreach $sysinfo_ref (@sys_list2) {
            if ($sysinfo_ref->{SSID} eq "hidden") {
                next;
            }
            if ($sysinfo_ref->{SSID} eq $tmp_sysinfo_ref->{SSID}) {
                #copy sysinfo hash into temporary fields
                foreach $hash_key (keys %{$sysinfo_ref}) {
                    $tmp_sysinfo_ref->{$hash_key} = $sysinfo_ref->{$hash_key};
                }
                $found = 1;
            }
        }
    }

    if ($tmp_sysinfo_ref->{HIDDEN} eq "yes") {
        $chk_ref->[0]->select;
    } else {
        $chk_ref->[0]->deselect;
    }
    if ($tmp_sysinfo_ref->{ENCRYPT} eq "off") {
        $rad_ref->[0]->select();
    } elsif ($tmp_sysinfo_ref->{ENCRYPT} eq "static") {
        $rad_ref->[1]->select();
    } elsif ($tmp_sysinfo_ref->{ENCRYPT} eq "auto") {
        $rad_ref->[2]->select();
    }
    if (defined $tmp_sysinfo_ref->{KEY}) {
        $ent_ref->[0]->delete(0, 'end');
        $ent_ref->[0]->insert('end', $tmp_sysinfo_ref->{KEY});
    }
    if (defined $tmp_sysinfo_ref->{CERTPATH}) {
        $ent_ref->[1]->delete(0, 'end');
        $ent_ref->[1]->insert('end', $tmp_sysinfo_ref->{CERTPATH});
    }
    if ($tmp_sysinfo_ref->{IPMODE} eq "dhcp") {
        $rad_ref->[3]->select();
    } else {
        $rad_ref->[4]->select();
    }
}
################################################################################
### UpdatePersysConfigInfo ($tmp_sysinfo_ref)
### Add/modify system configuration
### Input: $tmp_sysinfo_ref - configuration information to be deleted
### Global variables used: $conf_list_ref, $config_file_name, $interval,
###                        %ifc_config_list, @sys_list1, $sw1
################################################################################
sub UpdatePersysConfigInfo {
    my $tmp_sysinfo_ref = $_[0];
    my $conf_ref;
    my $ssid;
    my $hash_key;
    my $overwrite = 0;
    my %sysconf = ();
    my $item;
    my $hidden_ssid;

    print("UpdatePersysConfigInfo: hidden = $tmp_sysinfo_ref->{HIDDEN}\n");
    #Find existing configuration entry and modify it
    foreach $conf_ref (@{$conf_list_ref}) {
        $ssid = $conf_ref->{SSID};
        if ($conf_ref->{SSID} ne $tmp_sysinfo_ref->{SSID}) {
            next;
        } else {
            #found existing entry; overwrite it
            if ( ($conf_ref->{HIDDEN} eq 'no') && 
                 ($tmp_sysinfo_ref->{HIDDEN} eq 'no') ) {
                $item = 0;
                foreach $hidden_ssid (@hidden_list) {
                    if ($hidden_ssid eq $ssid) {
                        splice(@hidden_list, $item, 1);
                        last;
                    }
                    $item++;
                }
            }
            foreach $hash_key (keys %{$tmp_sysinfo_ref}) {
                $conf_ref->{$hash_key} = $tmp_sysinfo_ref->{$hash_key}
            }
            $overwrite = 1;
            print ("UpdatePersysConfigInfo: Entry found, overwrite\n");
        }
    }
    #No existing configuration entry found: add new entry
    unless ($overwrite) {
        print ("UpdatePersysConfigInfo: Entry not found, add new entry\n");
        foreach $hash_key (keys %{$tmp_sysinfo_ref}) {
            $sysconf{$hash_key} = $tmp_sysinfo_ref->{$hash_key};
        }
        #Change status from "Unconfigured" to "In Range"
        $sysconf{STATUS} = "In Range";
        if ($sysconf{HIDDEN} eq 'yes') {
            push(@hidden_list, $sysconf{SSID});
        }
        push (@{$conf_list_ref}, {%sysconf});
        unless (defined $ifc_config_list{$wlan_ifc}) {
            $ifc_config_list{$wlan_ifc} = $conf_list_ref;
        }
    }

    &WriteConfigFile($config_file_name, \$interval, \%ifc_config_list); 
    &UpdateDisplay();
    &AutomaticConnect(\@sys_list1, \@sys_list2);
    $sw1->withdraw;
    #Re-initialize configuration sub-window; activeperl bug countermeasure
    $sw1 = '';
}

################################################################################
### DeletePersysConfigInfo ($tmp_sysinfo_ref)
### Delete existing system configuration
### Input: $tmp_sysinfo_ref - configuration information to be deleted
################################################################################
#Delete existing system configuration
sub DeletePersysConfigInfo {
    my $tmp_sysinfo_ref = $_[0];
    my $conf_ref;
    my $ssid;
    my $delete = 0;
    my $item = 0;

    #Find existing configuration entry and modify it
    foreach $conf_ref (@{$conf_list_ref}) {
        $ssid = $conf_ref->{SSID};
        if ($conf_ref->{SSID} ne $tmp_sysinfo_ref->{SSID}) {
            #print("item $item, conf-ssid = $conf_ref->{SSID}, tmp-ssid = $tmp_sysinfo_ref->{SSID}\n");
        } else {
            #found existing entry; delete it
            #print("item $item, conf-ssid = $conf_ref->{SSID}, tmp-ssid = $tmp_sysinfo_ref->{SSID}\n");
            #print("###found###\n");
            $delete = 1;
            last;
        }
        $item++;
        next;
    }
    if ($delete) {
        splice(@{$conf_list_ref}, $item, 1);
        &WriteConfigFile($config_file_name, \$interval, \%ifc_config_list); 
        &UpdateDisplay();
        &AutomaticConnect(\@sys_list1, \@sys_list2);
    } else {
        #Trying to delete non-existing entry: do nothing
    }
    $sw1->withdraw;
    #Re-initialize configuration sub-window; activeperl bug countermeasure
    $sw1 = '';
}

################################################################################
### CancelPersysConfigInfo 
### Cancel per-system configuration and leave without changing
################################################################################
sub CancelPersysConfigInfo {
    $sw1->withdraw;
    #Re-initialize configuration sub-window; activeperl bug countermeasure
    $sw1 = '';
}

################################################################################
### OpenGeneralConfigWindow 
### Open General configuration window
### Callback: SaveGenConfigChange - on pressing 'OK' button
###           CancelGenConfigChange - on pressing 'Cancel' button
### Global variables used: $interval
################################################################################

sub OpenGeneralConfigWindow {
    my ($frame1, $frame2);

    #Open Sub-window
    if (Exists($sw2)) {
	return; 
    }
    $sw2 = $mw->Toplevel();
    $sw2 -> title("General Configuration");       

    #General Configuration
    $frame1 = $sw2->Frame(-borderwidth =>2, -relief => 'raised')
        ->pack(-side =>'top', -padx =>3, -pady =>3);
    $frame1 -> Label(-text => 'General Configuration ', -anchor=> 'nw')
        -> grid(-row => 0, -column  => 0, -sticky  => 'news');
    $frame1 -> Label(-text => "\tScan Interval", -anchor=> 'w')
        -> grid(-row => 1, -column  => 0, -sticky  => 'news');
    $frame1 -> Label(-text => "\tManagement Server Communication", -anchor=> 'w')
        -> grid(-row => 2, -column  => 0, -columnspan =>2, -sticky  => 'news');
    $frame1 -> Label(-text => "\tServer Address", -anchor=> 'w')
        -> grid(-row => 3, -column  => 0, -sticky  => 'news');
    $frame1 -> Label(-text => "Server Port", -anchor=> 'w')
        -> grid(-row => 3, -column  => 2, -sticky  => 'news');
    
    my @interval_list = ("0 (Disable Roaming)", "5 sec", "10 sec", "20 sec", "50 sec");
    my %tmp_geninfo;
    if ($interval == 0) {
        $tmp_geninfo{INTERVAL} = $interval_list[0];
    } else {
        $tmp_geninfo{INTERVAL} = ($interval/1000) . " sec";
    }
    $frame1 -> BrowseEntry(-variable=> \$tmp_geninfo{INTERVAL},
			   -choices => \@interval_list)
        -> grid(-row => 1, -column  => 1, -columnspan => 4, -sticky => 'news');

    $tmp_geninfo{SERVER} = $server_comm_mode;
    my $server_enable_button = $frame1 -> Radiobutton(-text => "enabled", 
						      -variable => \$tmp_geninfo{SERVER}, -value => "enabled")
        -> grid(-row => 2, -column  => 2, -sticky  => 'news');
    my $server_disable_button = $frame1 -> Radiobutton(-text => "disabled", 
						       -variable => \$tmp_geninfo{SERVER}, -value => "disabled")
        -> grid(-row => 2, -column  => 3, -sticky  => 'news');
    my $server_addr_entry = $frame1 -> Entry(-width => 15)
        -> grid(-row => 3, -column  => 1, -sticky  => 'news');
    my $server_port_entry = $frame1 -> Entry(-width => 10)
        -> grid(-row => 3, -column  => 3, -columnspan => 2, -sticky  => 'news');

    #OK/Cancel Buttons
    $frame2 = $sw2->Frame()
        ->pack(-side =>'right', -padx =>3, -pady =>3);
    $frame2 -> Button(-text => "Ok", -height => 1, -width =>8,
		      -command => [\&SaveGeneralConfigInfo, \%tmp_geninfo])
        ->pack(-side =>'left', -padx =>3, -pady =>3);
    $frame2 -> Button(-text => "Cancel", -height => 1, -width =>8,
		      -command => \&CancelGeneralConfigInfo)
        ->pack(-side =>'left', -padx =>3, -pady =>3);
    
}

################################################################################
### SaveGeneralConfigInfo ($tmp_geninfo_ref)
### Write down the value into configuration file, close the window and
###   re-initialize
### Input: $tmp_geninfo_ref - modified general configuration information
### Global variables used: $interval, $server_comm, $config_file_name,
###                        @sys_list1, \%ifc_config_list, \@hidden_list,
###                        $periodic_update_id
################################################################################
sub SaveGeneralConfigInfo {
    my ($tmp_geninfo_ref) = @_;
    my @words;
    my $tmp_interval;
    @words = split(/\s+/, $tmp_geninfo_ref->{INTERVAL});
    
    #Scan interval
    $tmp_interval = $words[0]*1000;
    if ($interval != $tmp_interval) {
        #Scan interval changed
        if ($periodic_update_id) {
            #Cancel future scan events
            $periodic_update_id->cancel;
        }
        $interval = $tmp_interval;
        if ($interval != 0) {
            #Schedule future periodic scan events
            $periodic_update_id = $mw->repeat($interval, \&PeriodicScan);
        }
    }
    #Server Communication
    $server_comm_mode = $tmp_geninfo_ref->{SERVER};

    &WriteConfigFile($config_file_name, \$interval, \%ifc_config_list); 
    &UpdateDisplay();
    &AutomaticConnect(\@sys_list1, \@sys_list2);
    $sw2->withdraw;
    #Re-initialize configuration sub-window; activeperl bug countermeasure
    $sw2 = '';
}

################################################################################
### CancelGeneralConfigInfo ($sw_ref)
### Cancel general configuration change and exit
### Global variables used: $sw2
################################################################################
sub CancelGeneralConfigInfo {
    #Leave without change
    $sw2->withdraw;
    #Re-initialize configuration sub-window
    $sw2 = '';
}


################################################################################
### Callback Functions (other)
################################################################################

################################################################################
### PeriodicScan
### Do periodic scanning of other cells in range
### Global variables used: $pending_update_id, @sys_list1
################################################################################
sub PeriodicScan {
    if ($pending_update_id) {
        #Cancel any pending scanning after automatic re-connect
        $pending_update_id->cancel;
    }
    #Scan other cells and update listbox
    &UpdateDisplay();
    #Handover to the best alternative system
    &AutomaticConnect(\@sys_list1, \@sys_list2);
}

################################################################################
### Functions for Wireless Extension Specific Information Retrieval
################################################################################

################################################################################
### GetIfcList
### Get Wireless LAN Interface List from /proc/net/wireless
### Return: @ifclist - Wireless LAN Interface List
################################################################################
sub GetIfcList {
    my $line_count;         
    my @words;              
    my @ifclist = ();       #Wireless LAN Interface List

    if ($linux) {
	open(WIRELESS_STAT, "/proc/net/wireless") || die "Can't open /proc/net/wireless\n";
    } else {
	open(WIRELESS_STAT, "proc_net_wireless.txt") || die "Can't open /proc/net/wireless\n";
    }
    for ($line_count = 0; <WIRELESS_STAT>; $line_count++) {
        if($line_count < 2){ next; }
        #delete left spaces
        $_ =~ s/^\s+//; 
        #Split into words
        @words = split(/\s+/, $_);
        #delete right colon
        chop($words[0]);
        #add interface list
        $ifclist[$line_count-2]=$words[0];
    }
    close(WIRELESS_STAT);
    if ($line_count == 2) {
        die "No Wireless Interface Available\n";
    }
    print "ifclist= @ifclist\n";
    return @ifclist;
}

################################################################################
### GetCurentApInfo($wlan_ifc, $current_ap_ref)
### Get Current Wireless LAN settings from iwconfig output
### Input: $wlan_ifc - wireless LAN interface name to be queried
### Input/Output: $current_ap_ref - reference to current wireless LAN
###                                 information
################################################################################
sub GetCurentApInfo {
    my ($ifc, $current_ap_ref) = @_;
    my @buf;            #iwconfig result buffer
    my ($line, @words, $element);
    my ($ssid, $tmp);
    my $prev_element = "";
    
    if ($linux) {
	#$elapsed = time()-$time_offset;
	#print ("###GetCurrentApInfo: Elapsed time = $elapsed [sec]\n");
	@buf = `/sbin/iwconfig $ifc`;
    } else {
	open(IWCONFIG_FILE, "iwconfig_eth1.txt");
	@buf = <IWCONFIG_FILE>;
	close(IWCONFIG_FILE);
    }

    foreach $line (@buf) {
        chomp $line;
        @words = split(/\s/, $line);
        foreach $element (@words) {
            if ($element =~ /ESSID/) {
                #Only one SSID can be associated
                ($tmp, $ssid) = split(/:/, $element);
                #Delete double quatation from SSID string
                $ssid =~ s/"//g; 
                $current_ap_ref->{SSID}  = $ssid;
            } elsif ($element =~ /Frequency/) {
                ($tmp, $current_ap_ref->{FREQ}) = split(/:/, $element);
            } elsif ($prev_element =~ /Point/) {
                $current_ap_ref->{BSSID} = $element;
            } elsif ($prev_element =~ /Signal/) {
                ($tmp, $current_ap_ref->{RSSI}) = split(/:/, $element);
            } 
            $prev_element = $element;
        }
    }
    print("GetCurentApInfo: ssid = $ssid, freq = $current_ap_ref->{FREQ}\n");
    print("                 bssid = $current_ap_ref->{BSSID}, rssi = $current_ap_ref->{RSSI}\n");
    if ($current_ap_ref->{BSSID} eq "44:44:44:44:44:44") {
        @buf = `cat /proc/net/wireless`;
        print("/proc/net/wireless\n");
        print("@buf");
    }


}

################################################################################
### GetApList ($ifc, $hidden_list_ref, $ap_list_ref)
### Get Wireless LAN information in range from iwlist scanning
### Input: $ifc - wireless LAN interface to scan
###        $hidden_list_ref - reference to the hidden SSID list
### Input/Output: $ap_list_ref - reference to the information of wireless LANs
###                              in range
################################################################################
sub GetApList {
    my ($ifc, $hidden_list_ref, $ap_list_ref) = @_;
    my @buf;
    my $hidden_ssid;

    #$elapsed = time() - $time_offset;
    #print("###GetApList: elapsed = $elapsed [sec]\n");
    #Initialize wireless LAN information in range
    @$ap_list_ref = ();

    #Scan for hidden systems
    foreach $hidden_ssid (@{$hidden_list_ref}) {
	if ($^O ne "MSWIN32") {
	    if($ifc =~ /ath/){
		#`/sbin/iwconfig $ifc essid $hidden_ssid`;
		@buf = `/sbin/iwlist $ifc scanning`; 
	    }
	    else {
		`/sbin/iwconfig $ifc essid ""`;
		@buf = `/sbin/iwlist $ifc scanning`;
		`/sbin/iwconfig $ifc essid $current_ap{SSID}`;
	    }
		open(IWLIST_FILE, "iwlist_eth1_scanning_eecs.txt");
		@buf = <IWLIST_FILE>;
	    close(IWLIST_FILE);
	}
	print("GetApList(ssid = $hidden_ssid), AddApList\n");
        AddApList(\@buf, $ap_list_ref);
    }
    #Scan for visible systems
    if ($^O ne "MSWIN32") {
	if($ifc =~ /ath/) {
	    #`/sbin/iwconfig $ifc essid ""`;
	    @buf = `/sbin/iwlist $ifc scanning`; 
	}
	else {
	    `/sbin/iwconfig $ifc essid ""`;
	    @buf = `/sbin/iwlist $ifc scanning`;
	    `/sbin/iwconfig $ifc essid $current_ap{SSID}`;
	}
	
    } else {
	open(IWLIST_FILE, "iwlist_eth1_scanning.txt");
	@buf = <IWLIST_FILE>;
	close(IWLIST_FILE);
    }
    print("GetApList(ssid = ANY), AddApList\n");
    AddApList(\@buf, $ap_list_ref);
}

################################################################################
### AddApList ($buf_ref, $ap_list_ref)
### Get Wireless LAN information in range from iwlist scanning
### Input: $buf_ref - reference to the iwlist result buffer
### Input/Output: $ap_list_ref - reference to the information of wireless LANs
###                              in range
################################################################################
sub AddApList {
    my ($buf_ref, $ap_list_ref) = @_;
    my ($item, $ssid, $element, $prev_element);
    my ($line, @words, $tmp);
    my %ap_info;

    $prev_element = "";
    $item = $#{$ap_list_ref} + 1;
    foreach $line (@{$buf_ref}) {
        chomp($line);
        @words = split(/ /, $line);
        foreach $element (@words) {
            if ($prev_element =~ /Address/) {
                $ap_list_ref->[$item]{BSSID} = $element;
                #print("bssid = $ap_list_ref->[$item]{BSSID}, ");
            } elsif ($element =~ /ESSID/) {
                #NOTE: Current "iwlist scanning" output can't handle    
                #      multiple SSIDs per AP (i.e. shared AP), therefore
                #      assume only one ssid output in iwlist result per line
                ($tmp, $ssid) = split(/:/, $element);
                #Delete double quatations at the end of SSID string
                $ssid =~ s/"//g;
                if ($ssid eq "") {
                    $ssid = "(hidden)";
                }
#"
                #print("ssid = $ssid, ");
                push (@{$ap_list_ref->[$item]{SSID}}, $ssid);
            } elsif ($element =~ /Frequency/) {
                ($tmp, $ap_list_ref->[$item]{FREQ}) = split(/:/, $element);
                #print("freq = $ap_list_ref->[$item]{FREQ}, ");
            } elsif ($prev_element =~ /Signal/) {
                ($tmp, $ap_list_ref->[$item]{RSSI}) = split(/:/, $element);
print("AddApList: signal = $ap_list_ref->[$item]{RSSI}, ");
            } elsif ($element =~ /key/) {
                ($tmp, $ap_list_ref->[$item]{ENCRYPT}) = split(/:/, $element);
print("AddApList: encrypt = $ap_list_ref->[$item]{ENCRYPT}\n");
                $item++;
            }
            $prev_element = $element;
        }
    }
}

################################################################################
### UpdateSysInfo ($clist_ref, $current_ap_ref, $ap_list_ref, $sys_list1_ref,
###                $sys_list2_ref) = @_;
### Update Wireless System Information for display
### Input: $clist_ref - reference to the configuration information hash
###        $current_ap_ref - reference to current wireless system information
###        $ap_list_ref - reference to the wireless system information in range
### Input/Output:
###        $sys_list1_ref - reference to the configured system list
###        $sys_list2_ref - reference to the unconfigured system list
################################################################################
sub UpdateSysInfo {
    my ($clist_ref, $current_ap_ref, $ap_list_ref, $sys_list1_ref, $sys_list2_ref) = @_;
    my @ap_list_2;      #Detected wireless system without duplicate entries
    my $conf_info;      #Configuration information element
    my %apnum = ();     #Hash table of number of access points per SSID
    my %apname = ();    #Hash table of one visible SSID per AP MAC address
    my %rssi = ();      #Hash table of received signal level per SSID
    my %config = ();    #Hash table of configured system per SSID
    my %sysinfo = ();   #Wireless system information for output
    my ($ap_ref, $conf_ref, $ssid_list_ref);
    my ($i, $ssid, $bssid, $encrypt);
    
    #Initialize configured/unconfigured wireless system information list
    @{$sys_list1_ref} = ();
    @{$sys_list2_ref} = ();

    #Make lookup tables and a list with no duplicate entries for later use
    foreach $ap_ref (@{$ap_list_ref}) {
        $ssid_list_ref = $ap_ref->{SSID};
        $bssid = $ap_ref->{BSSID};
        foreach $ssid (@{$ssid_list_ref}) {
            if ($ssid ne '(hidden)') {
                #One visible SSID for the AP
                $apname{$bssid} = $ssid;
            } elsif (!$apname{$bssid}) {
                #Hidden and visible SSID found;
                #Already counted by prior scan results
                next;
            }
            unless ($apnum{$ssid}) {
                #First visible SSID entry
                $apnum{$ssid} = 1;
                $rssi{$ssid} = $ap_ref->{RSSI};
                print("UpdateSysInfo: apnum($ssid) = $apnum{$ssid}, rssi = $rssi{$ssid}, freq = $ap_ref->{FREQ}\n");
                push(@ap_list_2, $ap_ref);      
            } else {
                #Count up AP numbers per SSID
                $apnum{$ssid}++;
                if ($ap_ref->{RSSI} > $rssi{$ssid}) {
                    $rssi{$ssid} = $ap_ref->{RSSI};
                }
                print("UpdateSysInfo: apnum($ssid) = $apnum{$ssid}, rssi = $ap_ref->{RSSI}, freq = $ap_ref->{FREQ}\n");
            }   
        }
    }
    
    #Configured Systems
    foreach $conf_ref (@{$clist_ref}) {
        %sysinfo = ();
        $ssid = $conf_ref->{SSID};
        unless ($config{$ssid}) {
            #lookup table for unconfigured systems
            $config{$ssid} = 1;
        }
        $sysinfo{SSID}  = $ssid;
        $sysinfo{ENCRYPT} = $conf_ref->{ENCRYPT};
        print("(UpdateSysInfo)ssid = $ssid, sysinfo{ENCRYPT} = $sysinfo{ENCRYPT}\n");
        if ($sysinfo{ENCRYPT} eq 'static') {
            $sysinfo{KEY} = $conf_ref->{KEY};
        }
        $sysinfo{IPMODE} = $conf_ref->{IPMODE};
        if ($sysinfo{IPMODE} eq 'static') {
            if (defined $conf_ref->{IPADDR}) {
                $sysinfo{IPADDR} = $conf_ref->{IPADDR};
            }
            if (defined $conf_ref->{IPMASK}) {
                $sysinfo{IPMASK} = $conf_ref->{IPMASK};
            }
            if (defined $conf_ref->{IPDGWY}) {
                $sysinfo{IPDGWY} = $conf_ref->{IPDGWY};
            }
            if (defined $conf_ref->{DNSPRI}) {
                $sysinfo{DNSPRI} = $conf_ref->{DNSPRI};
            }
            if (defined $conf_ref->{DNSSEC}) {
                $sysinfo{DNSSEC} = $conf_ref->{DNSSEC};
            }
        }
        $sysinfo{APNUM} = $apnum{$ssid};
        $sysinfo{HIDDEN} = $conf_ref->{HIDDEN};
        unless ($apnum{$ssid}) {
            $sysinfo{STATUS} = "Out of Range";
            $sysinfo{RSSI}       = &MinSignal;
        } elsif ($ssid eq $current_ap_ref->{SSID}) {
            $sysinfo{STATUS} = "Connected";
            $sysinfo{RSSI}       = $current_ap_ref->{RSSI};
        } else {
            $sysinfo{STATUS} = "In Range";
            $sysinfo{RSSI}       = $rssi{$ssid};
	    print("(UpdateSysInfo) ssid = $ssid, current_ap_ssid = $current_ap_ref->{SSID}\n");
        }
        print("(UpdateSysInfo) $ssid, STATUS = $sysinfo{STATUS}\n");
        push @{$sys_list1_ref}, {%sysinfo};
    }
    
    #Unconfigured Systems
    foreach $ap_ref (@ap_list_2) {
        $ssid_list_ref = $ap_ref->{SSID};
        foreach $ssid (@{$ssid_list_ref}) {
            unless ($config{$ssid}) {
                %sysinfo = ();
                $sysinfo{SSID} = $ssid;
                $sysinfo{ENCRYPT} = $ap_ref->{ENCRYPT};
                if ($sysinfo{ENCRYPT} eq 'on') {
                    $sysinfo{KEY} = "";
                }
                $sysinfo{APNUM} = $apnum{$ssid};
                $sysinfo{RSSI}  = $rssi{$ssid};
                $sysinfo{STATUS} = "Unconfigured";
		print("UpdateSysInfo(2): apnum($ssid) = $apnum{$ssid}, rssi = $ap_ref->{RSSI}, freq = $ap_ref->{FREQ}\n");
                push @{$sys_list2_ref}, {%sysinfo};
            }
        }
    }
}

################################################################################
### AutomaticConnect ($sys_list1_ref, $sys_list2_ref)
### Search available WLAN systems according to the priority and connect to the
###   first one
### Input: $sys_list1_ref - reference to the configured system list
### Input/Output: $ap_list_ref - reference to the information of wireless LANs
###                              in range
### Callback: UpdateDisplay - refresh main window after ($refresh_interval) msec
################################################################################
sub AutomaticConnect {
    my $sys_list1_ref = $_[0];
    my $sys_list2_ref = $_[1];
    my @buf;
    my $line;
    my $conn;
    my $pos = 0;
    my $tmp;
    my $sysinfo_ref;
    my $ssid;
    my $key;
    my $refresh_interval = 500;
    my $dhcp_cmd;
    my $dhcp_arg;
    my (@ipaddr, @ipmask, @bcast);
    my $bcast_adr;
    my $i;
    my $threshold = -85;

    foreach $sysinfo_ref (@{$sys_list2_ref}) {
            print("System 2: $sysinfo_ref->{SSID}\n");
            $sysinfo_ref->{STATUS} = "In Range";
            $sysinfo_ref->{ENCRYPT} = "auto";
            $sysinfo_ref->{IPMODE} = "dhcp";
	    #We make sure no key is set, because we do not know the information
	    # previously stored in the $sys_list2_ref table
	    $sysinfo_ref->{KEY} = "";

    	    #If we are connected to an unconfigured network, move it to the
            # first position within the unconfigured systems and mark it as
            # 'Connected'
    	    if ($sysinfo_ref->{SSID} eq $current_ap{SSID}) {
		$sysinfo_ref->{STATUS} = "Connected";
		$tmp = $sys_list2_ref->[0];
    	    	$sys_list2_ref->[0] = $sysinfo_ref;
            	$sys_list2_ref->[$pos] = $tmp;   
	    }

	    $pos++;
    }
    foreach $sysinfo_ref (@{$sys_list1_ref}, @{$sys_list2_ref}) {
        if ($sysinfo_ref->{STATUS} eq "Out of Range") {
            #The system is not available
            next;
        } elsif ($sysinfo_ref->{STATUS} eq "In Range"){ #&& $sysinfo_ref->{RSSI} > $threshold) {
            #Higher-priority system has become available: connect it
            $ssid = $sysinfo_ref->{SSID};
	    print("Execute handover to $ssid\n");
            my $time1 = gettimeofday();
                    my $encrypt = $sysinfo_ref->{ENCRYPT};
                    my $keytmp = $sysinfo_ref->{KEY};
		    print "Encryption: $encrypt\n";
		    print "Key: $keytmp\n";
	    if ($linux) {
                if ($sysinfo_ref->{ENCRYPT} eq "static") {
                    print "Static encryption\n";
		    if (defined $sysinfo_ref->{KEY}) {
                        $key = $sysinfo_ref->{KEY};
		        `/sbin/iwconfig $wlan_ifc key $key`;
		        `/sbin/iwconfig $wlan_ifc essid $ssid`;
                    }
                    else {
                        #Configuration error. Key must be specified
			print "Error: WEP key not configured\n";
                        next;
                    }
                } elsif ((!defined $sysinfo_ref->{KEY} || $sysinfo_ref->{KEY} eq '') &&
		         ($sysinfo_ref->{ENCRYPT} eq "off")) {
                    print "No encryption\n";
		    `/sbin/iwconfig $wlan_ifc encrypt off`;
		    `/sbin/iwconfig $wlan_ifc essid $ssid`;
		}

                elsif ((!defined $sysinfo_ref->{KEY} ||
                        $sysinfo_ref->{KEY} eq '') &&
                       ($sysinfo_ref->{ENCRYPT} eq 'auto')) {
                    print "802.1x authentication enabled\n";
                    my $xsupplicant_pid;
                    my $xsupplicant_arg;

                    #Kill old xsupplicant process, if any
                    $xsupplicant_pid = `/sbin/pidof -x xsupplicant`;
		    chomp ($xsupplicant_pid);
                    if ($xsupplicant_pid) {
                        #Kill xsupplicant process
                        `/bin/kill -9 $xsupplicant_pid`;
                        print "Killed old xsupplicant process $xsupplicant_pid\n";
                    }

		    #Clean the encryption key, so that, if xsupplicant fails,
		    # we do not connect via WEP using an old key
		    `/sbin/iwconfig $wlan_ifc key 0000-0000-0000-0000-0000-0000-00`;

                    #Start xsupplicant
		    `/sbin/iwconfig $wlan_ifc essid $ssid`;
                    $xsupplicant_arg = "-i $wlan_ifc -c $xsupplicant_conf";
                    #xsupplicant must be running during all the connection.
                    # I can not wait for it to finish
                    my $kidpid;
                    if (!defined($kidpid = fork())) {
                        die "cannot fork: $!";
                    } elsif ($kidpid == 0) {
                        exec("xsupplicant $xsupplicant_arg");
                        die "cannot exec xsupplicant: $!";
                    } else {
			print "Waiting for xsupplicant key\n";
                        my $l2_info = `$l2_info_extracter $wlan_ifc`;
                        my @l2_elements = split(" ", $l2_info);
                        $l2_key_digest = $l2_elements[0];
                        $mac_addr = $l2_elements[1];
                    }
                } else {
                    #Configuration error. Wrong encryption value, or 'auto' or
                    # 'off' plus specific key value.
		    print "Error: Wrong encryption parameters\n";
                    next;
                }

		print "Automatic Connect: IpMode = $sysinfo_ref->{IPMODE}\n";
	        #Check ISC DHCP client is running
	        my $dhcp_pid_path;
	        my $dhcp_pid;
	        $dhcp_pid_path = "/var/run/dhclient-$wlan_ifc.pid";
                $dhcp_pid = `/sbin/pidof -x dhclient`;
                chomp($dhcp_pid);
	        if ($dhcp_pid) {
		    #Release IP address
		    $dhcp_arg = "-1 -r -lf /var/lib/dhcp/dhclient-$wlan_ifc.leases -pf $dhcp_pid_path $wlan_ifc";
		    `/sbin/dhclient $dhcp_arg`;
		    `/sbin/iwconfig $wlan_ifc`;
		    #Kill DHCP client process
		    $dhcp_pid = `cat $dhcp_pid_path`;
		    kill $dhcp_pid;
                }
		if ($sysinfo_ref->{IPMODE} eq 'dhcp') {
			#Restart DHCP client
			$dhcp_arg = "-1 -lf /var/lib/dhcp/dhclient-$wlan_ifc -pf $dhcp_pid_path -cf /etc/dhclient-eth1.conf $wlan_ifc";
			@buf=`/sbin/dhclient $dhcp_arg 2>&1`;
		    #TODO: support for non-ISC DHCP clients (pump and dhcpcd)
		} else {         # ($sysinfo_ref->{IPMODE} eq 'static')
		    if (defined $sysinfo_ref->{IPADDR}) {
			`/sbin/ifconfig $wlan_ifc $sysinfo_ref->{IPADDR}`;
			if (defined $sysinfo_ref->{IPMASK}) {
			    `/sbin/ifconfig $wlan_ifc netmask $sysinfo_ref->{IPMASK}`;
			    @ipaddr = split(/\./, $sysinfo_ref->{IPADDR});
			    @ipmask = split(/\./, $sysinfo_ref->{IPMASK});
			    for ($i = 0; $i < 4; $i++) {
				$bcast[$i] = ($ipaddr[$i] | $ipmask[$i]) + (255-$ipmask[$i]);
			    }
			    $bcast_adr = "$bcast[0].$bcast[1].$bcast[2].$bcast[3]";
			    `/sbin/ifconfig $wlan_ifc broadcast $bcast_adr`;
			}
		    }
		    if (defined $sysinfo_ref->{IPDGWY}) {
			`/sbin/route add default gw $sysinfo_ref->{IPDGWY}`;
		    }
		    if (defined $sysinfo_ref->{DNSPRI}) {
			open(DNS_FILE, ">/etc/resolv.conf") ||
			    die "Can't open resolv.conf\n";
			#Write DNS server addresses
			print(DNS_FILE "nameserver $sysinfo_ref->{DNSPRI}\n");
			if (defined $sysinfo_ref->{DNSSEC}) {
			    print(DNS_FILE "nameserver $sysinfo_ref->{DNSSEC}\n");
			}
			close (DNS_FILE);
		    }
		}

		#Check if we have connectivity. If not, try the next network
		if ($sysinfo_ref->{IPMODE} eq 'dhcp') {
                    print("@buf");
		    $conn=1;
    		    foreach $line (@buf) {
        		chomp $line;
			print("$line\n");
                        if ($line eq 'No DHCPOFFERS received.') {
			    print("There is no connectivity\n");
			    $conn=0;
                        }
		    }
		    if (!$conn) {next;}
	        } else {
		    my $gw = $sysinfo_ref->{IPDGWY};
		    @buf=`ping $gw`;
		    $conn=1;
    		    foreach $line (@buf) {
        		chomp $line;
                        if (index($line,'0 received') != -1) {
			    print("There is no connectivity\n");
			    $conn=0;
                        }
		    }
		    if (!$conn) {next;}
		}
		 
		my $anp_res = $anpClient->login($l2_key_digest, $mac_addr);
                print "ANP Result: $anp_res\n";
                if ($anp_res ne 'Successful') {
                     print "Authentication failed. Roam to next network\n";
                     next;
                }

		$pending_update_id = $mw->after($refresh_interval, \&UpdateDisplay);
		my $time2 = gettimeofday();
                my $total = $time2 - $time1;
                print "total time: $total\n";

                #Just did handover to highest priority in range system:
                # stop seeking
                last;
	    }
	    
        } elsif ($sysinfo_ref->{STATUS} eq "Connected") {
            print("System connected\n");
            #Current System has the highest priority: stop seeking
            last;
        }
    }
}


################################################################################
### Main Window Components
################################################################################

################################################################################
### DisplayWlanSystems ($main_win, $sys_list1_ref, $sys_list2_ref, $list1_ref,
###                     $list2_ref)
### Display WLAN Interface List on Mainwindow
### Input: $main_win - main window
###        $sys_list1_ref - reference to configured system information
###        $sys_list2_ref - reference to unconfigured system information
###        $list1_ref - reference to configured system listboxes
###        $list2_ref - reference to unconfigured system listboxes
### Callback: MultiScrollSet1 - 1st listbox (configured system) scrolling
###           MultiScrollSet2 - 2nd listbox (unconfigured system) scrolling
###           GetSelectedSsid - on selecting items in the leftmost listbox
################################################################################
sub DisplayWlanSystems {
    my ($main_win, $sys_list1_ref, $sys_list2_ref, $list1_ref, $list2_ref) = @_;
    my $frame2;     #Mainwindow frame for displaying WLAN status listboxes
    my $yscrl1;     #Scrollbar widget for 1st listbox
    my $yscrl2;     #Scrollbar widget for 2nd listbox
    my @label = ('System ID', 'Signal', 'Status', '#AP');
    my $list1height = 6;            #Configured System Listbox height
    my @list1width = qw(22 4 12 3); #Configured System Listbox widths
    my $list2height = 4;            #Unconfigured System Listbox height
    my @list2width = qw(22 4 12 3); #Unconfigured System Listbox widths
    my $i;
    my $sysinfo_ref;

    $frame2 = $main_win->Frame
        ->pack(-side =>'left', -padx =>10, -pady =>0);
    
    for ($i = 0; $i < 4; $i++) {
        $frame2->Label(-text => $label[$i], -relief => 'raised')
            ->grid(-row => 0, -column => $i, -sticky => 'news');
    }
    $yscrl1 = $frame2->Scrollbar(-command => sub { &MultiScrollSet1; })
        ->grid(-row => 1, -column  => 4, -sticky  => 'news',);
    $yscrl2 = $frame2->Scrollbar(-command => sub { &MultiScrollSet2; })
        ->grid(-row => 2, -column  => 4, -sticky  => 'news',);
    
    for ($i = 0; $i < 4; $i++) {
        $list1_ref->[$i] = $frame2->Listbox(-height => $list1height, 
					    -width => $list1width[$i], 
					    -background => 'white',
					    -yscrollcommand   => ['set', $yscrl1])
            ->grid(-row => 1, -column => $i, -sticky => 'news');
        $list2_ref->[$i] = $frame2->Listbox(-height => $list2height, 
					    -width => $list2width[$i], 
					    -background => 'white',
					    -yscrollcommand   => ['set', $yscrl2])
            ->grid(-row => 2, -column => $i, -sticky => 'news');

    }

    foreach $sysinfo_ref (@{$sys_list1_ref}) {
        $list1_ref->[0]->insert('end', $sysinfo_ref->{SSID});
        $list1_ref->[1]->insert('end', &SignalStrength($sysinfo_ref->{RSSI}));
        $list1_ref->[2]->insert('end', $sysinfo_ref->{STATUS});
        $list1_ref->[3]->insert('end', $sysinfo_ref->{APNUM});
    }
    foreach $sysinfo_ref (@{$sys_list2_ref}) {
        $list2_ref->[0]->insert('end', $sysinfo_ref->{SSID});
        $list2_ref->[1]->insert('end', &SignalStrength($sysinfo_ref->{RSSI}));
        $list2_ref->[2]->insert('end', $sysinfo_ref->{STATUS});
        $list2_ref->[3]->insert('end', $sysinfo_ref->{APNUM});
    }
    $list1_ref->[0]->bind('<Button>', \&GetSelectedSsid);
    $list2_ref->[0]->bind('<Button>', \&GetSelectedSsid);
}

################################################################################
### ReDisplayWlanSystems ($sys_list1_ref, $sys_list2_ref, $list1_ref,
###                       $list2_ref, $selected_ssid_ref)
### Display WLAN Interface List on Mainwindow
### Input: $main_win - main window
###        $sys_list1_ref - reference to configured system information
###        $sys_list2_ref - reference to unconfigured system information
###        $list1_ref - reference to configured system listboxes
###        $list2_ref - reference to unconfigured system listboxes
###        $selected_ssid_ref - reference to the SSID selected in the Listbox
### Callback: MultiScrollSet1 - 1st listbox (configured system) scrolling
###           MultiScrollSet2 - 2nd listbox (unconfigured system) scrolling
###           GetSelectedSsid - on selecting items in the leftmost listbox
################################################################################
sub ReDisplayWlanSystems {
    my ($sys_list1_ref, $sys_list2_ref, $list1_ref, $list2_ref, $selected_ssid_ref) = @_;
    my ($sysinfo_ref, $i, $size);

    #remove old entries from the list
    for ($i = 0; $i <= 3; $i++ ) {
        $size = $list1_ref->[$i]->size;
        $list1_ref->[$i]->delete(0,$size-1);
        $size = $list2_ref->[$i]->size;
        $list2_ref->[$i]->delete(0,$size-1);
    }
    #Insert configured system entries
    foreach $sysinfo_ref (@{$sys_list1_ref}) {
        $list1_ref->[0]->insert('end', $sysinfo_ref->{SSID});
        if ( ($selected_ssid ne '') &&
             ($sysinfo_ref->{SSID} eq $selected_ssid) ) {
            $list1_ref->[0]->selectionSet('end');
        }
        $list1_ref->[1]->insert('end', &SignalStrength($sysinfo_ref->{RSSI}));
        $list1_ref->[2]->insert('end', $sysinfo_ref->{STATUS});
        $list1_ref->[3]->insert('end', $sysinfo_ref->{APNUM});
    }
    #Insert unconfigured system entries
    foreach $sysinfo_ref (@{$sys_list2_ref}) {
        $list2_ref->[0]->insert('end', $sysinfo_ref->{SSID});
        if ( ($selected_ssid ne '') &&
             ($sysinfo_ref->{SSID} eq $selected_ssid) ) {
            $list2_ref->[0]->selectionSet('end');
        }
        $list2_ref->[1]->insert('end', &SignalStrength($sysinfo_ref->{RSSI}));
        $list2_ref->[2]->insert('end', $sysinfo_ref->{STATUS});
        $list2_ref->[3]->insert('end', $sysinfo_ref->{APNUM});
    }
}


################################################################################
### DisplayWlanIfList ($main_win, $wlan_ifc_list_ref, $selected_wlan_ifc_ref)
### Display WLAN Interface List on Mainwindow
### Input: $main_win - main window
###        $wlan_ifc_list_ref - reference to the WLAN interface list
###        $wlan_ifc_ref - reference to the selected WLAN interface
### Callback: ChangeIfc - on selecting a WLAN interface
################################################################################
sub DisplayWlanIfList {
    my ($main_win, $wlan_ifc_list_ref, $wlan_ifc_ref) = @_;
    my $frame1;         #Mainwindow frame for WLAN interface listing
    my $ifcdisp;        #BrowseEntry for WLAN interface list
    my $selected_wlan_ifc;  #Selected WLAN interface

    $selected_wlan_ifc = $$wlan_ifc_ref;
    $frame1 = $main_win->Frame
        ->pack(-anchor=>'w', -padx => 10, -pady => 10);
    $ifcdisp = $frame1->BrowseEntry(-label   => "Interface",
                                    -variable=> \$selected_wlan_ifc,
                                    -choices => $wlan_ifc_list_ref,
                                    -browsecmd => [\&ChangeIfc, \$selected_wlan_ifc])
        ->pack(-side =>'left');
    return;
}

################################################################################
### DisplayButtons ($main_win, $wlan_ifc_list_ref, $selected_wlan_ifc_ref)
### Display WLAN Interface List on Mainwindow
### Input: $main_win - main window
### Callback: PriorityUp - on pressing 'arrow_up' button 
###           PriorityDown - on pressing 'arrow_down' button
###           OpenPersysConfigWindow - on pressing 'per-system configuration'
###                                    button
###           OpenGeneralConfigWindow - on pressing 'general configuration'
###                                     button
################################################################################
sub DisplayButtons {
    my ($main_win, $sub_win1, $sub_win2) = @_;
    my $frame3;         #Mainwindow frame for priority/configuration buttons
    my $pri_label;      #Label for prioirty change buttons
    my $config_label;   #Label for configuration change buttons

    $frame3 = $main_win->Frame
        ->pack(-side =>'left', -padx =>10, -pady =>10);

    #Display Buttons for changing priority
    $pri_label = $frame3->Label(-text => "Priority")
        ->pack(-anchor=>'w', -side =>'top', -padx =>20, -pady =>3);
    $frame3->Button(-bitmap => '@arrow_up.xbm',
		    -height => 30, -width =>50,
		    -command => \&PriorityUp)
        ->pack(-anchor=>'w', -side =>'top', -padx =>18, -pady =>0);
    $frame3->Button(-bitmap => '@arrow_down.xbm',
		    -height => 30, -width =>50,
		    -command => \&PriorityDown)
        ->pack(-anchor=>'w', -side =>'top', -padx =>18, -pady =>0);
    #Display Buttons for configuration
    $config_label = $frame3->Label(-text => "Configuration")
        ->pack(-anchor=>'w', -side =>'top', -padx =>0, -pady =>5);
    $frame3->Button(-text => "Per-System",
		    -height => 1, -width =>10,
		    -command => \&OpenPersysConfigWindow)
        ->pack(-anchor=>'w', -side =>'top', -padx =>0, -pady =>0);
    $frame3->Button(-text => "General",
		    -height => 1, -width =>10,
		    -command => \&OpenGeneralConfigWindow)
        ->pack(-anchor=>'w', -side =>'top', -padx =>0, -pady =>5);
}

################################################################################
### SignalStrength ($rssi)
### Translate relative signal strength to numer of asterisks
### Input: $rssi - measured RSSI value
### Return: {"","*","**","***","****","*****"}
################################################################################
sub SignalStrength {
    my $rssi = $_[0];
    my $min_signal = &MinSignal();
    if ($rssi == $min_signal) {
        return "";
    } elsif ($rssi <= -82) {
        #802.11a 6Mbps Sensitivity Limit
        return "*";
    } elsif ($rssi <= -77) {
        #802.11a 18Mbps Sensitivity Limit
        return "**";
    } elsif ($rssi <= -70) {
        #802.11a 36Mbps Sensitivity Limit
        return "***";
    } elsif ($rssi <= -65) {
        #802.11a 54Mbps Sensitivity Limit
        return "****";
    } else {
        return "*****";
    }
}

################################################################################
### MinSignal ()
### Define minimum signal level
### Return: -255
################################################################################
sub MinSignal {
    return -255;
}

################################################################################
### Configuration File Read/Write Routines               
################################################################################

################################################################################
### GetConfigFileName (@argv)
### Get configuration file name from command line argument
### Input: @argv - command line argument
### Return: $filename
################################################################################
sub GetConfigFileName {
    print "Getting configuration file name...\n";
    my @argv = $_[0];
    my $argc = @argv;
    my $filename;
    
    if ($argc > 2) {
        die "Usage: wrh.pl <config file>\n";
    }
    if ($argc > 1) {
        $filename = $argv[1];
        if ($filename eq 'help') {
            die "Usage: wrh.pl <config file>\n";
        }
        unless (-e $filename) {
            die "Can't find configuration file $filename\n";
        }
    } else {
        $filename = $ROAMING_CLIENT_CONFIG_FILE_PATH;
        unless (-e $filename) {
            $filename = "";
            print "No configuration file. Defaults will be used\n";
        }
    }
    return ($filename);
}
################################################################################
### ReadConfigFile($filename, $interval_ref, $ifc_config_list_ref,
###                $hidden_list_ref)
### Read Configuration Data from File
### Input: $filename - configuration file name
### Input/Output: $interval_ref - reference to $interval (scanning interval)
###               $ifc_config_list_ref - reference to $ifc_config_list (system
###                                      configuration per interface)
###               $hidden_list_ref - reference to $hidden_list (hidden SSID list)
################################################################################
sub ReadConfigFile {
    print "Reading configuration file...\n";
    my ($filename, $interval_ref, $ifc_config_list_ref, $hidden_list_ref) = @_;
    my $ifc;
    my %sysconf;
    my @syslist;
    my @words;
    my ($tmp, $bssid, $ssid);
    my @ssid_list;
    
    open(CONFIG_FILE, $filename) ||
        die "Can't open config file $filename\n";

    while (<CONFIG_FILE>) {
        if ($_ =~ /#/) {
            next;
        } elsif ($_ =~ /Interval/) {
            #delete CR/LF
            chomp($_);
            $_ =~ s/^\s+//; 
            ($tmp, $$interval_ref) = split(/ /, $_);
            print("ReadConfigFile: interval = $$interval_ref\n");
        } elsif ($_ =~ /Interface/) {
            @words = split(/ /, $_);
            $ifc = $words[1];
            print "ReadConfigFile: ifc = $ifc\n";
            @syslist = ();
            while (<CONFIG_FILE>) {
                if ($_ =~ /#/) {
                    next;
                } elsif ($_ =~ /System/) {
                    %sysconf = ();
                    while (<CONFIG_FILE>) {
                        if ($_ =~ /#/) {
                            next;
                        } elsif ($_ =~ /SSID/) {
                            chomp($_);
                            @words = split(/ /, $_);
                            $sysconf{SSID} = $words[1];
                            print "\tssid = $sysconf{SSID}, ";
                        } elsif ($_ =~ /Hidden/) {
                            chomp($_);
                            @words = split(/ /, $_);
                            $sysconf{HIDDEN} = $words[1];
                            print "\thidden = $sysconf{HIDDEN}, ";
                        } elsif ($_ =~ /Encryption/) {
                            chomp($_);
                            @words = split(/ /, $_);
                            $sysconf{ENCRYPT} = $words[1];
                            #print "encrypt = $sysconf{ENCRYPT}, ";
                        } elsif ($_ =~ /Key/) {
                            chomp($_);
                            @words = split(/ /, $_);
                            $words[1] =~ s/-//g;
                            $sysconf{KEY} = $words[1];
                            print "key = $sysconf{KEY}\n";
                        } elsif ($_ =~ /CertPath/) {
                            chomp($_);
                            ($tmp, $sysconf{CERTPATH}) = split(/ /, $_);
                            print "CertPath = $sysconf{CERTPATH}\n";
                        } elsif ($_ =~ /IpMode/) {
                            chomp($_);
                            ($tmp, $sysconf{IPMODE}) = split(/ /, $_);
                            print "\tIpMode = $sysconf{IPMODE}\n";
                        } elsif ($_ =~ /IpAddr/) {
                            chomp($_);
                            ($tmp, $sysconf{IPADDR}) = split(/ /, $_);
			    print "ReadConfigFile: IpAddr = $sysconf{IPADDR}\n";
                        } elsif ($_ =~ /IpMask/) {
                            chomp($_);
                            ($tmp, $sysconf{IPMASK}) = split(/ /, $_);
                        } elsif ($_ =~ /IpDgwy/) {
                            chomp($_);
                            ($tmp, $sysconf{IPDGWY}) = split(/ /, $_);
                        } elsif ($_ =~ /DnsPri/) {
                            chomp($_);
                            ($tmp, $sysconf{DNSPRI}) = split(/ /, $_);
                        } elsif ($_ =~ /DnsSec/) {
                            chomp($_);
                            ($tmp, $sysconf{DNSSEC}) = split(/ /, $_);
                        } elsif ($_ =~ /}/) {
			#End of system description
			if ($sysconf{HIDDEN} eq 'yes') {
			    push(@{$hidden_list_ref}, $sysconf{SSID});
			}
			push(@syslist, {%sysconf});
			print "\tend(system)\n";
			last;
		    } else {
			die ("Error: mal-formed config file(system)\n");
		    }
		}
	    } elsif ($_ =~ /}/) {
	    #End of interface description
	    $ifc_config_list_ref->{$ifc} = [@syslist];
	    print "end(interface)\n";
	    last;
	} else {
	    die "Error: mal-formed config file(interface)\n";
	}
    }
}
}
close(CONFIG_FILE);
return;
}

################################################################################
### WriteConfigFile($filename, $interval_ref, $ifc_config_list_ref)
### Write Configuration Data to File
### Input: $filename - configuration file name
### Input/Output: $interval_ref - reference to $interval (scanning interval)
###               $ifc_config_list_ref - reference to $ifc_config_list (system configuration per interface)
################################################################################
sub WriteConfigFile {
    my ($filename, $interval_ref, $ifc_config_list_ref) = @_;
    my ($ifc, $sysconf_ref, $ssid_list_ref);
    my ($bssid, $ssid);
    my $backup_file_name;

    #Backup old configuration file
    $backup_file_name = $filename . ".bak";

    if (-e $backup_file_name) {
        unlink($backup_file_name);
    }
    rename ($filename, $backup_file_name);
    
    open(CONFIG_FILE, ">$filename") ||
        die "Can't open config file $config_file_name\n";
    
    #Write periodic update interval time
    print(CONFIG_FILE "#Periodic Update Interval\n");
    print(CONFIG_FILE "Interval $$interval_ref\n");
    print(CONFIG_FILE "\n");
    
    #Write per-interface configuration 
    print(CONFIG_FILE "#Interface Configuration\n");
    foreach $ifc (keys %{$ifc_config_list_ref}) {
        print(CONFIG_FILE "Interface $ifc {\n");
        foreach $sysconf_ref (@{$ifc_config_list_ref->{$ifc}}) {
            print(CONFIG_FILE "\tSystem {\n");
            print(CONFIG_FILE "\t\tSSID $sysconf_ref->{SSID}\n");
            print(CONFIG_FILE "\t\tHidden $sysconf_ref->{HIDDEN}\n");
            print(CONFIG_FILE "\t\tEncryption $sysconf_ref->{ENCRYPT}\n");
            if (defined $sysconf_ref->{KEY}) {
                print(CONFIG_FILE "\t\tKey $sysconf_ref->{KEY}\n");
            }
            if (defined $sysconf_ref->{CERTPATH}) {
                print(CONFIG_FILE "\t\tCertPath $sysconf_ref->{CERTPATH}\n");
            }
            print(CONFIG_FILE "\t\tIpMode $sysconf_ref->{IPMODE}\n");
            if ($sysconf_ref->{IPMODE} eq 'static') {
                if (defined $sysconf_ref->{IPADDR}) {
                    print(CONFIG_FILE "\t\tIpAddr $sysconf_ref->{IPADDR}\n");
                }
                if (defined $sysconf_ref->{IPMASK}) {
                    print(CONFIG_FILE "\t\tIpMask $sysconf_ref->{IPMASK}\n");
                }
                if (defined $sysconf_ref->{IPDGWY}) {
                    print(CONFIG_FILE "\t\tIpDgwy $sysconf_ref->{IPDGWY}\n");
                }
                if (defined $sysconf_ref->{DNSPRI}) {
                    print(CONFIG_FILE "\t\tDnsPri $sysconf_ref->{DNSPRI}\n");
                }
                if (defined $sysconf_ref->{DNSSEC}) {
                    print(CONFIG_FILE "\t\tDnsSec $sysconf_ref->{DNSSEC}\n");
                }
            }
            print(CONFIG_FILE "\t}\n");
        }
        print(CONFIG_FILE "}\n");
    }
    print(CONFIG_FILE "\n");
    
    close(CONFIG_FILE);
}
