/*
 * $Id: ArtifactManager.java,v 1.5.2.1 2002/09/17 01:52:20 jsergent Exp $
 * ArtifactManager.java
 *
 *  The contents of this file are subject to the Sun Industry Standards
 *  Source License Version 1.1 (the "License"); You may not use this file
 *  except in compliance with the License. You may obtain a copy of the
 *  License at http://www.experimentalstuff.com/Technologies/IPL/download.html.
 *
 *  Software provided under this License is provided on an "AS IS" basis,
 *  WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING,
 *  WITHOUT LIMITATION, WARRANTIES THAT THE SOFTWARE IS FREE OF DEFECTS,
 *  MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE, OR NON-INFRINGING.
 *  See the License for the specific provisions governing your rights and
 *  obligations concerning the Software.
 *
 *  The Initial Developer of the Original Code is: Sun Microsystems, Inc.
 *
 *  Copyright: 2002 by Sun Microsystems, Inc.
 *
 *  All Rights Reserved.
 *
 *  Contributor(s): _______________________________________
 */

package com.sun.liberty;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

/**
 * ArtifactManager provides artifact registration services.
 * It can be used to register a message and its artifact for future
 * dereferencing.  It also stores the intended audience of the artifact
 * so that message signature checking can be performed before retrieving
 * and invalidating the artifact.
 * 
 * @author	Jonathan Sergent
 * @version	$Revision: 1.5.2.1 $
 */

public class ArtifactManager {
    /** Map from artifact to message object. */
    static Map messageMap = new HashMap();
    /** Map from artifact to intended audience string (provider ID) */
    static Map intendedAudienceMap = new HashMap();
    /** Map from artifact to date object indicating insertion time */
    static Map ageMap = new HashMap();

    /**
     * Registers an artifact-to-message mapping.  Also stores an
     * intended audience, which should be the provider ID of the Liberty
     * provider who will be allowed to dereference the artifact.  Only
     * requests signed by this provider's key will successfully dereference
     * the artifact.  (This is enforced by the servlet processing the
     * request.)
     *
     * @param artifact The artifact (must be globally unique)
     * @param msg The message
     * @param intendedAudience The provider ID of the provider who may look up
     *        the message.
     */
    static void registerArtifact(byte[] artifact, LibertyArtifactMessage msg, String intendedAudience) {
	if (Config.debug) {
	    System.err.println("registerArtifact artifact:" + artifact);
	    try {
		System.err.println("b64:" + URLUtils.base64Encode(artifact));
	    } catch (Exception e) {}
	}
	String s = new String(artifact);
	synchronized (ArtifactManager.class) {
	    messageMap.put(s, msg);
	    intendedAudienceMap.put(s, intendedAudience);
	    ageMap.put(s, new Date());
	}
    }

    /**
     * Retrieves the message associated with the given artifact.
     * The message may only be retrieved once.  If the message is too
     * old, or if it cannot be found, it will return null.
     *
     * @param artifact The artifact to retrieve
     * @return The message, or null if the operation failed.
     */
    static LibertyArtifactMessage retrieveArtifact(byte[] artifact) {
	if (Config.debug) {
	    System.err.println("retrieveArtifact artifact:" + artifact);
	    try {
		System.err.println("b64:" + URLUtils.base64Encode(artifact));
	    } catch (Exception e) {}
	}
	LibertyArtifactMessage msg;
	String s = new String(artifact);
	long age = new Date().getTime();
	synchronized (ArtifactManager.class) {
	    age -= ((Date) ageMap.get(s)).getTime();
	    msg = (LibertyArtifactMessage) messageMap.get(s);
	    if (Config.debug) {
		System.err.println("retrieveArtifact: msg:" + msg);
	    }
	    messageMap.remove(s);
	    intendedAudienceMap.remove(s);
	    ageMap.remove(s);
	}
	if (age < Config.artifactAgeLimit) {
	    if (msg == null) {
		// XXX: Should have a better way to log these things.
		System.err.println("WARNING: Failed artifact query");
	    }
	    return msg;
	} else {
	    System.err.println("WARNING: Expired artifact query");
	    return null;
	}
    }

    /**
     * Looks up the intended audience for an artifact without
     * invalidating it.  This can be used to check the signature on
     * an artifact request before actually invalidating the artifact,
     * so that a third party who has captured the artifact but not
     * the key of the intended audience cannot deny service by invalidating
     * the artifact.
     *
     * @param artifact The artifact
     * @return The provider ID of the intended audience of the artifact.
     */
    public static String retrieveAudience(byte[] artifact) {
	synchronized (ArtifactManager.class) {
	    return (String) intendedAudienceMap.get(new String(artifact));
	}
    }
}
