/*
 * $Id: LogoutNotificationProcessor.java,v 1.4.2.1 2002/09/17 01:52:25 jsergent Exp $
 * LogoutNotificationProcessor.java
 *
 *  The contents of this file are subject to the Sun Industry Standards
 *  Source License Version 1.1 (the "License"); You may not use this file
 *  except in compliance with the License. You may obtain a copy of the
 *  License at http://www.experimentalstuff.com/Technologies/IPL/download.html.
 *
 *  Software provided under this License is provided on an "AS IS" basis,
 *  WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING,
 *  WITHOUT LIMITATION, WARRANTIES THAT THE SOFTWARE IS FREE OF DEFECTS,
 *  MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE, OR NON-INFRINGING.
 *  See the License for the specific provisions governing your rights and
 *  obligations concerning the Software.
 *
 *  The Initial Developer of the Original Code is: Sun Microsystems, Inc.
 *
 *  Copyright: 2002 by Sun Microsystems, Inc.
 *
 *  All Rights Reserved.
 *
 *  Contributor(s): _______________________________________
 */

package com.sun.liberty;
import com.sun.liberty.saml.SAMLNameIdentifier;
import java.util.Iterator;
import java.util.Set;

/**
 * Processes LogoutNotifications.  Note that this is only useful for
 * back-channel logout notifications.  Front channel logout is too
 * complex for this model, and is implemented completely in a servlet.
 *
 * @author Jonathan Sergent
 * @version $Revision: 1.4.2.1 $
 */
public class LogoutNotificationProcessor implements RequestProcessor {
    private static ProviderDescriptorFactory dir;

    static {
        try {
            dir = ProviderDescriptorFactory.getInstance();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Processes the LogoutNotification.  Indicates an error if a SOAP fault
     * will be necessary.
     *
     * @param request The notification request
     * @return a RequestProcessorAction, indicating RETURN_RESPONSE if
     * processing was successful or FAILED if a SOAP fault or other error
     * should be returned to the client
     */
    public RequestProcessorAction process(LibertyLogoutNotification request) throws Exception
    {
        String signedBy = request.getSignedBy();
        if (signedBy == null) {
            if (Config.debug) { System.err.println("Logout isn't signed"); }
            return RequestProcessorAction.FAILED;
        }
        String providerID = request.getProviderID();
        if (providerID == null) {
            if (Config.debug) { System.err.println("Logout missing prov ID"); }
            return RequestProcessorAction.FAILED;
        }
        if (!providerID.equals(signedBy)) {
            if (Config.debug) { System.err.println("Logout signed wrong"); }
            return RequestProcessorAction.FAILED;
        }
        ProviderDescriptor prov = dir.find(providerID);
        if (prov == null) {
            /* signature check should have failed, but... */
            if (Config.debug) { System.err.println("Can't find provider?"); }
            return RequestProcessorAction.FAILED;
        }

        SAMLNameIdentifier nameID = request.getNameIdentifier();

        LibertyUser requestedUser = null;

        if (prov instanceof SPDescriptor) {
            requestedUser = LibertyUser.find((SPDescriptor) prov, nameID);
        }

        if (requestedUser == null && prov instanceof IDPDescriptor) {
            requestedUser = LibertyUser.find((IDPDescriptor) prov, nameID);
        }

        if (requestedUser == null) {
            /* Federation was probably already terminated */
            if (Config.debug) { System.err.println("Can't find user"); }
            return RequestProcessorAction.FAILED;
        }

        // XXX: should be getCurrentSession(request.getSessionIndex())
        LibertyUserSession fcSession;
        fcSession = (LibertyUserSession) requestedUser.getCurrentSession();
        if (Config.debug) {
            System.err.println("requestedUser:" + requestedUser);
            System.err.println("fcSession:" + fcSession);
        }
        if (fcSession == null) {
            return RequestProcessorAction.FAILED;
        }
        Set sps = fcSession.getRelyingSPList();
	if (sps != null) {
	    if (Config.debug) {
		System.err.println("Found relying SPs");
	    }
	    Iterator i = sps.iterator();
	    while (i.hasNext()) {
		SPDescriptor sp = (SPDescriptor) i.next();
	 	if (sp.getProviderID().equals(providerID)) {
		    // Do not notify sender
		    continue;
		}
		if (sp.getSingleLogoutProtocolProfile() !=
		    LibertyProfile.SL_IDP_SOAP)
		{
		    return RequestProcessorAction.FAILED;
		}
		SAMLNameIdentifier nameid = requestedUser.getNameProvidedBy(sp);
		LibertySoapLogoutNotification n;
		n = new LibertySoapLogoutNotification();
		n.setNameIdentifier(nameid);
		n.setProviderID(Config.localProviderID);
		n.setSignedBy(Config.localProviderID);
		SoapUtils.send(n, sp.getSoapEndpoint().toString());
	    }
            sps.clear();
	} else {
	    if (Config.debug) {
		System.err.println("No relying SPs");
	    }
	}
        fcSession.setUser((LibertyUser) null);
	if (Config.debug) {
	    System.err.println("User logged out via incoming SOAP call");
	}

        return RequestProcessorAction.RETURN_RESPONSE;
    }
}
