/*
 * $Id: SAMLAssertion.java,v 1.12.2.2 2002/09/17 01:52:29 jsergent Exp $
 * SAMLAssertion.java
 *
 *  The contents of this file are subject to the Sun Industry Standards
 *  Source License Version 1.1 (the "License"); You may not use this file
 *  except in compliance with the License. You may obtain a copy of the
 *  License at http://www.experimentalstuff.com/Technologies/IPL/download.html.
 *
 *  Software provided under this License is provided on an "AS IS" basis,
 *  WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING,
 *  WITHOUT LIMITATION, WARRANTIES THAT THE SOFTWARE IS FREE OF DEFECTS,
 *  MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE, OR NON-INFRINGING.
 *  See the License for the specific provisions governing your rights and
 *  obligations concerning the Software.
 *
 *  The Initial Developer of the Original Code is: Sun Microsystems, Inc.
 *
 *  Copyright: 2002 by Sun Microsystems, Inc.
 *
 *  All Rights Reserved.
 *
 *  Contributor(s): _______________________________________
 */

package com.sun.liberty.saml;
import com.sun.liberty.Config;
import com.sun.liberty.LibertyAuthenticationStatement;
import com.sun.liberty.XMLUtils;
import java.util.Date;
import java.util.List;
import java.util.LinkedList;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Data bindings for saml:Assertion element.
 * Get and set methods correspond to the child elements of Assertion
 * in the XML Schema.  getStatements and setStatements take as an argument
 * a List containing SAMLStatement objects or subclasses thereof.
 *
 * @author Jonathan Sergent
 * @version $Revision: 1.12.2.2 $
 */

public class SAMLAssertion extends SAMLElement {
    private int majorVersion = 1;
    private int minorVersion = 0;
    private String assertionID = generateIDType();
    private String issuer = null;
    private Date issueInstant = new Date();
    private SAMLConditions conditions = null;
    private SAMLAdvice advice = null;
    private boolean signed = false;
    private String signedBy = null;
    private List statements = null;


    public void parse(Element e) {
	setMajorVersion(Integer.valueOf(e.getAttributeNS(null, "MajorVersion")).intValue());
	setMinorVersion(Integer.valueOf(e.getAttributeNS(null, "MinorVersion")).intValue());
	setAssertionID(e.getAttributeNS(null, "AssertionID"));
	setIssuer(e.getAttributeNS(null, "Issuer"));
	setIssueInstant(parseDate(e.getAttributeNS(null, "IssueInstant")));

	NodeList list = e.getChildNodes();
	Element signature = null;
	for (int j = 0; j < list.getLength(); j++) {
	    if (list.item(j) instanceof org.w3c.dom.Text) { continue; }
	    Element child = (Element) list.item(j);
	    SAMLStatement statementToAdd = null;
	    if (isElement(child, SAML_NS_URI, "Conditions")) {
		SAMLConditions c = new SAMLConditions();
		c.parse(child);
		setConditions(c);
	    } else if (isElement(child, SAML_NS_URI, "Advice")) {
		SAMLAdvice c = new SAMLAdvice();
		c.parse(child);
		setAdvice(c);
	    } else if (isElement(child, SAML_NS_URI, "AuthenticationStatement")) {
		String t = child.getAttributeNS(XSI_NS_URI, "type");
		if (t != null && t.equals("lib:AuthenticationStatementType")) {
		    statementToAdd = new LibertyAuthenticationStatement();
		} else {
		    statementToAdd = new SAMLAuthenticationStatement();
		}
	    } else if (isElement(child, SAML_NS_URI, "AuthorizationDecisionStatement")) {
		statementToAdd = new SAMLAuthorizationDecisionStatement();
	    } else if (isElement(child, SAML_NS_URI, "AttributeStatement")) {
		statementToAdd = new SAMLAttributeStatement();
	    } else if (isElement(child, DSIG_NS_URI, "Signature")) {
		signature = child;
		if (Config.debug) {
	            System.err.println("SAMLAssertion: found signature");
		    System.err.println("signature:" + signature);
		}
	    }
	
	    if (statementToAdd != null) {
		statementToAdd.parse(child);
		List statements = getStatements();
		if (statements == null) {
		    statements = new LinkedList();
		}
		statements.add(statementToAdd);
		setStatements(statements);
	    }
	}
	if (signature != null) {
	    if (Config.debug) {
		System.err.println("SAMLAssertion: verifying");
		System.err.println("issuer:" + issuer);
		System.err.println("signature:" + signature);
	    }
	    setSignedBy(XMLUtils.verify(signature, issuer) ? issuer : null);
	} else {
	    System.err.println("SAMLAssertion: didn't find signature");
	}
    }

    public Element getElement(Document doc) {
        Element a = doc.createElementNS(SAML_NS_URI, SAML_NS_ID + ":Assertion");
        addAttributesTo(a);
        addChildrenTo(a);
        return a;
    }

    protected void addAttributesTo(Element a) {
        a.setAttributeNS(null, "MajorVersion", new Integer(majorVersion).toString());
        a.setAttributeNS(null, "MinorVersion", new Integer(minorVersion).toString());
        a.setAttributeNS(null, "AssertionID", assertionID);
	if (issuer != null) {
	    a.setAttributeNS(null, "Issuer", issuer);
	}
        a.setAttributeNS(null, "IssueInstant", formatDate(issueInstant));
    }

    protected void addChildrenTo(Element a) {
        if (conditions != null) {
            a.appendChild(conditions.getElement(a.getOwnerDocument()));
        }
        if (advice != null) {
            a.appendChild(advice.getElement(a.getOwnerDocument()));
        }
        iterateAddChildren(a, statements);
        if (signedBy != null) {
            XMLUtils.sign(a);
        }
    }
    

    public int getMajorVersion() {
        return majorVersion;
    }
    
    public void setMajorVersion(int v) {
        majorVersion = v;
    }

    public int getMinorVersion() {
        return minorVersion;
    }

    public void setMinorVersion(int v) {
        minorVersion = v;
    }

    public String getAssertionID() {
        return assertionID;
    }

    public void setAssertionID(String id) {
        assertionID = id;
    }

    public Date getIssueInstant() {
	return issueInstant;
    }

    public void setIssueInstant(Date value) {
	issueInstant = value;
    }

    public String getIssuer() {
        return issuer;
    }

    public void setIssuer(String name) {
        issuer = name;
    }

    public SAMLConditions getConditions() {
        return conditions;
    }

    public void setConditions(SAMLConditions value) {
        conditions = value;
    }

    public SAMLAdvice getAdvice() {
        return advice;
    }

    public void setAdvice(SAMLAdvice value) {
        advice = value;
    }

    public String getSignedBy() {
        return signedBy;
    }

    public void setSignedBy(String id) {
        signedBy = id;
    }

    public List getStatements() {
        return statements;
    }

    public void setStatements(List value) {
        statements = value;
    }
}
