/*
 * $Id: AuthFilter.java,v 1.1.2.2 2002/09/17 01:52:33 jsergent Exp $
 * AuthFilter.java
 *
 *  The contents of this file are subject to the Sun Industry Standards
 *  Source License Version 1.1 (the "License"); You may not use this file
 *  except in compliance with the License. You may obtain a copy of the
 *  License at http://www.experimentalstuff.com/Technologies/IPL/download.html.
 *
 *  Software provided under this License is provided on an "AS IS" basis,
 *  WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING,
 *  WITHOUT LIMITATION, WARRANTIES THAT THE SOFTWARE IS FREE OF DEFECTS,
 *  MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE, OR NON-INFRINGING.
 *  See the License for the specific provisions governing your rights and
 *  obligations concerning the Software.
 *
 *  The Initial Developer of the Original Code is: Sun Microsystems, Inc.
 *
 *  Copyright: 2002 by Sun Microsystems, Inc.
 *
 *  All Rights Reserved.
 *
 *  Contributor(s): Ana Sanz Merino [University of California Berkeley, 2003]
 */

package com.sun.liberty.servlets;

import com.sun.liberty.Config;
import com.sun.liberty.LibertyUser;
import com.sun.liberty.LibertyUserSession;
import javax.servlet.*;
import javax.servlet.http.*;

/**
 * Servlet filter which forwards request to LoginServlet unless user's
 * HttpSession is logged in.  If user is logged in, the filter
 * wraps the request to modify getRemoteUser() to return the username
 * retrieved from the LibertyUser object.
 *
 * @author Jonathan Sergent
 * @version $Revision: 1.1.2.2 $
 */

public class AuthFilter implements Filter {
    FilterConfig config;

    public void init(FilterConfig config) {
	this.config = config;
    }

    public void destroy() {}

    public void doFilter(ServletRequest request, ServletResponse response,
	    FilterChain chain) throws java.io.IOException, ServletException {

	if (!(request instanceof HttpServletRequest)) {
	    /*
	     * This filter relies on the HttpSession to maintain user
	     * session state, so it won't work with non-HTTP requests.
	     */
	    chain.doFilter(request, response);
	    return;
	}

	/* Get session. */
	HttpServletRequest req = (HttpServletRequest) request;
	HttpSession ses = req.getSession(true);
	LibertyUserSession lus;
	lus = (LibertyUserSession) ses.getAttribute("liberty.session");
	if (lus == null) {
	    if (Config.debug) {
		System.err.println("AuthFilter: Creating new session.");
	    }
	    lus = new LibertyUserSession(); 
	    ses.setAttribute("liberty.session", lus);
	}
	else if (Config.debug) {
	    System.err.println("AuthFilter: Using existing session.");
	}

	String reqUrl = req.getRequestURL().toString();
        if (Config.debug) {
            System.err.println("AuthFilter: Requested URL: " + reqUrl);
        }
	LibertyUser lu = lus.getUser();
	boolean blocked = false;
	if (lu == null) {
	    if (Config.debug) {
		System.err.println("AuthFilter: No user found.");
	    }

            /* If no Liberty authenticated, block access to login pages, 
             * so that Login Servlet takes control and Liberty authentication 
             * can be done. This must be done for both http and https.
             */
            if ((reqUrl.equals(Config.localProviderApp + Config.loginPage)) ||
                (reqUrl.equals(Config.localProviderApp + 
                        Config.autoLoginPage)) ||
                (reqUrl.equals(Config.alternativeLocalProviderApp + 
                        Config.loginPage)) ||
                (reqUrl.equals(Config.alternativeLocalProviderApp + 
                        Config.autoLoginPage))) {
                blocked = true;
            } 

            /* If no Liberty authenticated, block access to pages adapted 
             * to particular users and which are exclusive for Liberty users.
             * This must be done for both http and https.
             */
            else if ((reqUrl.equals(Config.localProviderApp +
                        Config.federationPage)) ||
                (reqUrl.equals(Config.localProviderApp + 
                        Config.successfulLibertyAuthnPage)) ||
                (reqUrl.equals(Config.localProviderApp +
                        Config.alreadyLibertyAuthnPage)) ||
                (reqUrl.equals(Config.alternativeLocalProviderApp +
                        Config.federationPage)) ||
                (reqUrl.equals(Config.alternativeLocalProviderApp + 
                        Config.successfulLibertyAuthnPage)) ||
                (reqUrl.equals(Config.alternativeLocalProviderApp +
                        Config.alreadyLibertyAuthnPage))) {
		blocked = true;
            }

            /* If no Liberty authenticated, block access to rest of pages only
             * if no Radius authenticated either
             */
	    else {
                if (ses.getAttribute("radius.authenticated") == null) {
	            if (Config.debug) {
		        System.err.println("AuthFilter: No radius authenticated.");
	            }
                    blocked = true;
                }
            }
        }
            
	/* Forward request to Login Servlet if access to the page is blocked. */
        if (blocked) {
	    ServletContext context = config.getServletContext();
	    RequestDispatcher disp =
		context.getNamedDispatcher("LoginServlet");
	    if (Config.debug) {
		System.err.println("AuthFilter: Forwarding to LoginServlet.");
	    }
	    disp.forward(request, response);
	    return;
	} else {
            /* If a Liberty authenticated user requests one the login pages, 
             * indicate him that he has already been authenticated.
             * It has no sense for him to access the login page while Liberty
             * authenticated, as it is the authentication option with associated
             * access to all the available of pages.
             * If the user wants to authenticate again, he must logout first.
             * This must be done for both http and https.
             */
            if ((lu != null) && 
                ((reqUrl.equals(Config.localProviderApp + Config.loginPage)) ||
                 (reqUrl.equals(Config.localProviderApp +
                         Config.autoLoginPage)) ||
                 (reqUrl.equals(Config.alternativeLocalProviderApp +
                         Config.loginPage)) ||
                 (reqUrl.equals(Config.alternativeLocalProviderApp +
                         Config.autoLoginPage)))) {
	        ServletContext context = config.getServletContext();
	        RequestDispatcher disp = 
                    context.getRequestDispatcher(
                        Config.alreadyLibertyAuthnPage);
	        if (Config.debug) {
		    System.err.println("AuthFilter: Forwarding to Already Authenticated Page.");
	        }
	        disp.forward(request, response);
	        return;
            }

	    /* For the rest of non-blocked pages, allow filtered request to get
             * through.
             */
	    String temp = null;
	    try {
		temp = lu.getUsername();
	    } catch (Exception e) {}
	    final String user = temp;
	    if (Config.debug) {
		System.err.println("AuthFilter: Logged-in user name:" + user);
	    }
	    HttpServletRequest wrapped = new HttpServletRequestWrapper(req) {
		public String getRemoteUser() {
		    return user;
		}
	    };
	    /*
	     * XXX: Should we also wrap getSubject()?  Probably.
	     * Should we call doAs()?  Probably not.
	     */
	    chain.doFilter(wrapped, response);
	    return;
	}
    }
}
