package ucb.eecs.sahara.sso.anp;

import ucb.eecs.sahara.sso.DebugUtils;

import com.sun.liberty.Config;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import java.util.List;
import java.util.Iterator;

import ucb.eecs.sahara.sso.sp.AuthnControlSession;
import ucb.eecs.sahara.sso.anp.saml.SAMLPQuery;
import ucb.eecs.sahara.sso.authn.FirewallController;


/**
 * @author Ana Sanz Merino [University of California Berkeley, 2003]
 * @version $Revision: 1.1.1.1 $
 */

public class SimpleANPRequestProcessor implements ANPRequestProcessor {    

    private static final String localMethod = "local";
    private static final String radiusMethod = "radius";
    private static final String libertyMethod = "liberty";

    private static final boolean iAccessValue = true;

    private static AuthnControlSession getControlSession(HttpServletRequest req) {
        HttpSession ses = req.getSession(true);
        AuthnControlSession ctls;
        ctls = (AuthnControlSession) ses.getAttribute("control.session");
        if (ctls == null) {
            ctls = new AuthnControlSession();
            ses.setAttribute("control.session", ctls);
        }
        return ctls;
    }

    public ANPRequestProcessorAction process(HttpServletRequest httpReq,
                                             HttpServletResponse httpResp,
					     ANPRequest request,
					     ANPRequestProcResult procResult) 
    {
        DebugUtils.debugTime("SimpleANPRequestProcessor", "Entering process",
                             DebugUtils.LOW_LEVEL_FUNC);

        try {
            /* Check version */
            if (request.getMajorVersion() != 1) {
                if (Config.debug) {
                    System.err.println(
                        "rejected due to major version mismatch");
                }
                procResult.failureCode = "samlp:VersionMismatch";
                return ANPRequestProcessorAction.RETURN_RESPONSE;
            }

            /* Retrieve query */
            SAMLPQuery query = request.getQuery();
            if (query == null) {
                if (Config.debug) {
                    System.err.println("Got ANPRequest without query");
                }
                procResult.failureCode = "samlp:Sender";
                return ANPRequestProcessorAction.RETURN_RESPONSE;
            }

            DebugUtils.debugTime("SimpleANPRequestProcessor", "Proceeding to process query",
                                 DebugUtils.LOW_LEVEL_FUNC);

            /* ANP Authentication Capabilities query processing */
            if (query instanceof ANPAuthnCapabilitiesQuery) {
                ANPAuthnCapabilitiesQuery authnCapabQuery = (ANPAuthnCapabilitiesQuery) query;
                ANPAuthnCapabQueryResult acqr = new ANPAuthnCapabQueryResult();
                acqr.nonce = authnCapabQuery.getNonce();

                procResult.queryType = ANPRequestProcResult.AUTHN_CAPB_QUERY;
                procResult.queryResult = acqr;
                return ANPRequestProcessorAction.RETURN_RESPONSE;
            }

            /* ANP Authentication query processing */
            if (query instanceof ANPAuthnQuery) {
                ANPAuthnQuery authnQuery = (ANPAuthnQuery) query;

                /* Check authentication method is included */
                String authnMethod = authnQuery.getAuthnMethodIDReference();
                if (authnMethod == null) {
                    if (Config.debug) {
                        System.err.println("Got ANPAuthnQuery without ANPAuthnMethodIDReference");
                    }
                    procResult.failureCode = "samlp:Sender";
                    return ANPRequestProcessorAction.RETURN_RESPONSE;
                }

                /* Check charging option is included */
                String chargingOption = 
                    authnQuery.getChargingOptionIDReference();
                if (chargingOption == null) {
                    if (Config.debug) {
                        System.err.println("Got ANPAuthnQuery without ANPChargingOptionIDReference");
                    }
                    procResult.failureCode = "samlp:Sender";
                    return ANPRequestProcessorAction.RETURN_RESPONSE;
                }

                /* Prepare to check authentication method ad charging option are
                 * supported.
                 */
                Iterator i = null;
                boolean found;
                ANPAuthnCapabilitiesStatement my_acs =
                    ANPAuthnCapabilitiesManager.getAuthnCapabilitiesStatement();
                DebugUtils.debugTime("SimpleANPRequestProcessor",
                                     "Proceeding to check authnMethod & ChargOpt are supported",
                                     DebugUtils.LOW_LEVEL_FUNC);

                /* Check authentication method is supported */
                ANPAuthnMethod am = null;
                found = false;
                List supportedAuthnMethods = my_acs.getAuthnMethods();
                i = supportedAuthnMethods.iterator();
                while (i.hasNext()) {
                    am = (ANPAuthnMethod) i.next();
                    if (authnMethod.equals(am.getAuthnMethodID())) {
                        found = true;
                        break;
                    }
                }
                if (!found) {
                    if (Config.debug) {
                        System.err.println("ANPAuthnMethodIDReference indicates an authentication method not supported");
                    }
                    procResult.failureCode = "samlp:Sender";
                    return ANPRequestProcessorAction.RETURN_RESPONSE;
                }
                 
                /* Check charging option is supported */
                ANPChargingOption co = null;
                found = false;
                List supportedChargingOptions = my_acs.getChargingOptions();
                i = supportedChargingOptions.iterator();
                while (i.hasNext()) {
                    co = (ANPChargingOption) i.next();
                    if (chargingOption.equals(co.getChargingOptionID())) {
                        found = true;
                        break;
                    }
                }
                if (!found) {
                    if (Config.debug) {
                        System.err.println("ANPChargingOptionIDReference indicates a charging option not supported");
                    }
                    procResult.failureCode = "samlp:Sender";
                    return ANPRequestProcessorAction.RETURN_RESPONSE;
                }

                /* Optional: Check all required user information is included in the query */

                /* Authenticate the user */
                DebugUtils.debugTime("SimpleANPRequestProcessor",
                                     "Proceeding to authenticate user",
                                     DebugUtils.LOW_LEVEL_FUNC);

                ANPAuthnQueryProcessor aqp = null;
                ANPRequestProcessorAction action = null;
                ANPAuthnQueryProcResult aqProcResult = 
                    new ANPAuthnQueryProcResult();

                if (authnMethod.equals(localMethod)) {
                    aqp = new SimpleLocalANPAuthnQueryProcessor();
                } else if (authnMethod.equals(radiusMethod)) {
                    aqp = new SimpleRadiusANPAuthnQueryProcessor();
                } else if (authnMethod.equals(libertyMethod)) {
                    aqp = new SimpleLibertyANPAuthnQueryProcessor();
                } else {
                    if (Config.debug) {
                        System.err.println(
                            "Not possible to process the ANPRequest");
                    }
                    return ANPRequestProcessorAction.FAILED;
                }

                action = aqp.process(httpReq, httpResp, authnQuery,
                                     aqProcResult);

                DebugUtils.debugTime("SimpleANPRequestProcessor",
                                     "Authentication finished",
                                     DebugUtils.LOW_LEVEL_FUNC);

                return processAuthnQueryProcResult(httpReq, httpResp,
                                                   authnQuery, procResult,
                                                   authnMethod, chargingOption,
                                                   iAccessValue,
                                                   action, aqProcResult); 
            }

            /* ANP Authentication continuation query processing */
            if (query instanceof ANPAuthnContQuery) {
                ANPAuthnContQuery authnContQuery = (ANPAuthnContQuery) query;
                AuthnControlSession ctls = getControlSession(httpReq);

                /* Check authentication method is included in the HttpSession */
                String authnMethod = ctls.getAuthnMethod();
                if (authnMethod == null) {
                    if (Config.debug) {
                        System.err.println("Got ANPAuthnContQuery without authn method");
                    }
                    return ANPRequestProcessorAction.FAILED;
                }

                /* Check charging option is included in the HttpSession */
                String chargingOption = ctls.getChargingOption();
                if (chargingOption == null) {
                    if (Config.debug) {
                        System.err.println("Got ANPAuthnContQuery without charging option");
                    }
                    return ANPRequestProcessorAction.FAILED;
                }

                /* Look for the external access information in the HttpSession.
                 * If it was not set, the default will be 'false'
                 */
                boolean externalAccess = ctls.getExternalAccess();

                /* Process the request */
                ANPAuthnContQueryProcessor aqp = null;
                ANPRequestProcessorAction action = null;
                ANPAuthnQueryProcResult aqProcResult = 
                    new ANPAuthnQueryProcResult();

                if (authnMethod.equals(libertyMethod)) {
                    aqp = new SimpleLibertyANPAuthnContQueryProcessor();
                } else {
                    if (Config.debug) {
                        System.err.println(
                            "Not possible to process the ANPRequest");
                    }
                    return ANPRequestProcessorAction.FAILED;
                }

                action = aqp.process(httpReq, httpResp, authnContQuery,
                                     aqProcResult);

                DebugUtils.debugTime("SimpleANPRequestProcessor",
                                     "Authentication continuation processing finished",
                                     DebugUtils.LOW_LEVEL_FUNC);

                /* Process the result of processing the request */
                return processAuthnQueryProcResult(httpReq, httpResp,
                                                   authnContQuery, procResult,
						   authnMethod, chargingOption,
                                                   externalAccess,
                                                   action, aqProcResult); 
            } else {
                if (Config.debug) {
                    System.err.println("Got ANPRequest with unknown query");
                }
                procResult.failureCode = "samlp:Sender";
                return ANPRequestProcessorAction.RETURN_RESPONSE;
            }
        } catch (Exception e) {
            if (Config.debug) {
                System.err.println(e);
            }
            return ANPRequestProcessorAction.FAILED;
        }
    }


    /* Processes the result of processing an ANPAuthnQuery or ANPAuthnContQuery
     * The function is shared by both types of ANP queries because the response
     *  returned to the client for both types of queries is exactly the same.
     * For both types of queries, the same query type is indicated as result to 
     *  the invoker of the SimpleANPRequestProcessor, so that it continues 
     *  treating both queries in the same way.
     */
    private ANPRequestProcessorAction processAuthnQueryProcResult(
                                            HttpServletRequest httpReq,
                                            HttpServletResponse httpResp,
                                            SAMLPQuery query,
					    ANPRequestProcResult procResult, 
                                            String authnMethod,
                                            String chargingOption,
                                            boolean externalAccess,
					    ANPRequestProcessorAction action, 
                                            ANPAuthnQueryProcResult aqProcResult)
                                      throws Exception
    {
        ANPRequestProcessorAction closeFwAction = null;

	if (action == ANPRequestProcessorAction.RETURN_RESPONSE) {
	    /* Only if no error, the ANP Authentication Statement must
	     * be returned
	     */
	    if (aqProcResult.failureCode == null) { //If no error, 
		if (aqProcResult.authnQueryResult != null) {
		    if (Config.debug) {
			System.err.println("AuthnQueryResult: " + aqProcResult.authnQueryResult.authnResult);
		    }
		    if (aqProcResult.authnQueryResult.authnResult.equals("Successful")) {
			/* Validate L2 info */
			if (Config.debug) {
			    System.err.println("Let's validate L2 info");
			}

			ANPAuthnQueryL2Processor l2processor = null;
			ANPRequestProcessorAction l2action = null; 
			ANPAuthnQueryProcResult l2procResult = 
			    new ANPAuthnQueryProcResult();

			l2processor = new SimpleANPAuthnQueryL2Processor();
			l2action = l2processor.process(httpReq, httpResp, query,
						       l2procResult);

			if (l2action == ANPRequestProcessorAction.RETURN_RESPONSE) {
			    if (l2procResult.failureCode == null) { //If no error, 
				if (l2procResult.authnQueryResult != null) {
				    if (l2procResult.authnQueryResult.authnResult.equals("Successful")) {
                                        /* If there was previously a redirection, close the 
                                           hole previously open in firewall for redirection */
                                        if (query instanceof ANPAuthnContQuery) {
                                            closeFwAction = closeRedirectionHole(httpReq, aqProcResult);
                                            if (closeFwAction.equals(ANPRequestProcessorAction.FAILED)) {
			                        return ANPRequestProcessorAction.FAILED;
                                            }
                                            /* Wait so that there are not race conditions with opening firewall */
                                            java.lang.Thread.sleep(20);
                                        }

					/* If requested, allow user to access the external network */
					if (externalAccess) {
					    FirewallController.giveAccess(httpReq);
					}
				    }

                                    procResult.queryType = ANPRequestProcResult.AUTHN_QUERY;
                                    procResult.queryResult = l2procResult.authnQueryResult;
                                    return ANPRequestProcessorAction.RETURN_RESPONSE;
				} else {
				    return ANPRequestProcessorAction.FAILED;
				}
			    } else {
				procResult.failureCode = l2procResult.failureCode;
				return ANPRequestProcessorAction.RETURN_RESPONSE;
			    }
			} else {
			    return ANPRequestProcessorAction.FAILED;
			} 
		    } else if (aqProcResult.authnQueryResult.authnResult.equals("Redirect")) {
			/* Store control information to be used when the authentication
			 * continues at this server */
			AuthnControlSession ctls = getControlSession(httpReq);
			ctls.setExternalAccess(externalAccess);
			ctls.setAuthnMethod(authnMethod);
			ctls.setChargingOption(chargingOption);

                        /* If there was previously a redirection, close the 
                           hole previously open in firewall for redirection */
                        if (query instanceof ANPAuthnContQuery) {
                            closeFwAction = closeRedirectionHole(httpReq, aqProcResult);
                            if (closeFwAction.equals(ANPRequestProcessorAction.FAILED)) {
			        return ANPRequestProcessorAction.FAILED;
                            }
                        }

			/* Allow user to access the redirect URL */
			FirewallController.giveURLAccess(httpReq,
			    aqProcResult.authnQueryResult.redirectServer);

		        procResult.queryType = ANPRequestProcResult.AUTHN_QUERY;
		        procResult.queryResult = aqProcResult.authnQueryResult;
                        return ANPRequestProcessorAction.RETURN_RESPONSE;
		    } else { //authnResult equals "Failed"
		        procResult.queryType = ANPRequestProcResult.AUTHN_QUERY;
		        procResult.queryResult = aqProcResult.authnQueryResult;
		        return ANPRequestProcessorAction.RETURN_RESPONSE;
                    }
		} else {
		    return ANPRequestProcessorAction.FAILED;
		}
	    } else {
		procResult.failureCode = aqProcResult.failureCode;
		return ANPRequestProcessorAction.RETURN_RESPONSE;
	    }
	} else {
	    return ANPRequestProcessorAction.FAILED;
        } 
    }

    /* Close the hole previously open in firewall for redirection */
    private ANPRequestProcessorAction closeRedirectionHole(
                                            HttpServletRequest httpReq,
                                            ANPAuthnQueryProcResult aqProcResult)
                                      throws Exception
    {
        String redirectedServer = aqProcResult.authnQueryResult.redirectedServer;
        if (redirectedServer == null) {
	    if (Config.debug) {
	        System.err.println("Got ANPAuthnContQuery without redirected server");
	    }
	    return ANPRequestProcessorAction.FAILED;
	}

	FirewallController.removeURLAccess(httpReq, redirectedServer);
        return ANPRequestProcessorAction.RETURN_RESPONSE;
    }
}
